/*
 * drivers/linux2.4/fndiag.c
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#include <kernel_config_stuff.h>

#define __NO_VERSION__		/* don't define kernel_version in module.h */
#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>	/* printk() */
#include <linux/slab.h>	/* kmalloc() */
#include <linux/vmalloc.h>
#include <linux/errno.h>	/* error codes */
#include <linux/types.h>	/* size_t */
#include <linux/proc_fs.h>
#include <linux/pci.h>

#include <asm/uaccess.h>	/* put_user */
#include <asm/io.h>

#include <qikdata_hba.h>	/* qd hardware definitions */
#include <qikdata_drive.h>	/* local definitions */
#include <linux/blk.h>

void
qikdata_dump_gendisk(struct gendisk *gdev, kdev_t dev, unsigned minors,
    struct file_operations *ops, long size)
{
#ifdef QIKDATA_DEBUG
    printk(KERN_ALERT HERE "qikdata_dump_gendisk(struct gendisk *gdev,"
	" kdev_t dev, unsigned minors,\n                    "
	"        struct file_operations *ops, long size)\n\n");

    printk(KERN_ALERT HERE "struct gendisk *gdev     = 0x%08lX\n",
	(u_long)gdev);
    printk(KERN_ALERT HERE "kdev_t dev               = 0x%04X\n", dev);
    printk(KERN_ALERT HERE "unsigned minors          = 0x%04X\n", minors);
    printk(KERN_ALERT HERE "struct file_operations\n");
    printk(KERN_ALERT HERE "                *ops     = 0x%08lX\n",
	(u_long)ops);
    printk(KERN_ALERT HERE "long size                = %ld (0x%08lX)\n",
	size, size);
    printk(KERN_ALERT HERE "gendisk structure: \n");
    printk(KERN_ALERT HERE "   int  major             = %d (%04X)\n",
	gdev->major, gdev->major);
    printk(KERN_ALERT HERE "   const char *major_name = %s\n",
	gdev->major_name);
    printk(KERN_ALERT HERE "   int minor_shift        = %d\n",
	gdev->minor_shift);
    printk(KERN_ALERT HERE "   int max_p              = %d\n", gdev->max_p);
    printk(KERN_ALERT HERE "   struct hd_struct *part = 0x%08lX\n",
	(u_long) gdev->part);
    printk(KERN_ALERT HERE "   int *sizes             = 0x%08lX\n",
	(u_long) gdev->sizes);
    printk(KERN_ALERT HERE "   int nr_real            = %d\n", gdev->nr_real);
    printk(KERN_ALERT HERE " \n");
    printk(KERN_ALERT HERE " \n");
    printk(KERN_ALERT HERE " MINOR(dev) >> gdev->minor_shift = 0x%04X\n",
	(MINOR(dev) >> gdev->minor_shift));
#endif
}
