/*
 * drivers/linux2.4/fnexec.c
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#include <kernel_config_stuff.h>

#define __NO_VERSION__		/* don't define kernel_verion in module.h */
#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>	/* HDEBUG() */
#include <linux/slab.h>	/* kmalloc() */
#include <linux/vmalloc.h>
#include <linux/errno.h>	/* error codes */
#include <linux/types.h>	/* size_t */

#include <asm/uaccess.h>	/* put_user */
#include <asm/delay.h>
#include <asm/io.h>

#include <qikdata_hba.h>	/* qd hardware definitions */
#include <qikdata_drive.h>	/* local definitions */
#include <fndimm.h>

#include <debug_misc.h>
#include <fnexec.h>

/*#define RW_DEBUG KERN_ERR */
#define RW_DEBUG KERN_ALERT
/*#define RW_DEBUG KERN_DEBUG */
/*#define SGL_DEBUG KERN_ERR */
#define SGL_DEBUG KERN_INFO
/*#define SGL_DEBUG KERN_DEBUG */

#define QD_READ  1
#define QD_WRITE 0

/* quick fudge for now */
extern int qikdata_hardsect;

void
qikdata_link_reset(kdev_t i_rdev)
{
    PHW_DEVICE_EXTENSION devExt = &qikdata_device_extension[QD_HBA_NR(i_rdev)];
    u_int   pRTR = (u_int) devExt->RTRmap;

    writel(readl(pRTR + REG_IO) & ~IO_LINK_RESET, pRTR + REG_IO);
    udelay(10);
    writel(readl(pRTR + REG_IO) | IO_LINK_RESET, pRTR + REG_IO);
    udelay(50);
    writel(readl(pRTR + REG_IO) & ~IO_LINK_RESET, pRTR + REG_IO);
    udelay(20);

    /*
     * Clear ECC bits
     */
    writel(1, pRTR + REG_STATUS_HIGH);
    writel(1, pRTR + REG_ECC);

}


static int
Handle_DMA(int rdwt, kdev_t i_rdev, u_char * buffer, int sector,
    int nr_sectors)
{
    PHW_DEVICE_EXTENSION devExt = &qikdata_device_extension[QD_HBA_NR(i_rdev)];
    u_int   pRTR = (u_int) devExt->RTRmap;
    PUNIT_DETAILS unitDet = & (devExt->UnitDet[QD_DRV_NR(i_rdev)]);
    u_long  locPhysAddr;
    u64     locLinAddr;
    u64     physicalAddress;
    SGL_ENTRY Sgl;
    u64     size;
    u64     bytes_to_send;
    int     dimm;
    int     bytes_available;
    int     bytes_needed;
    int     n_cols;
    int     finished = 0;

    HDEBUG(KERN_ALERT HERE "Handle_DMA(rdwt=%d, i_rdev=%d, buffer=%p, "
	"sector=%d, nr_secords=%d)\n", rdwt, (int)i_rdev, buffer, sector,
	nr_sectors);
    size = (u64) nr_sectors *(u64) qikdata_hardsect;
    locLinAddr = (u64) sector *(u64) qikdata_hardsect;

    /*
     * clear the fifo and any interrupts
     */
    writel(DMA_FIFO_CLEAR, pRTR + REG_DMA);
    writel(DMA_SGL_INTERRUPT | DMA_ABORT_INTERRUPT, pRTR + REG_DMA);
    while (!finished)
    {
	unsigned long dma;

	locPhysAddr =
	    qikdata_addr2lb
	    (
		unitDet->drive_address,
		unitDet->Dimm,
		locLinAddr,
		&Sgl
	    );
	HDEBUG(KERN_ALERT HERE "Logical = %08lX %08lX, Physical = %08lX\n",
	    (u_long) (locLinAddr >> 32), (u_long) (locLinAddr & 0xFFFFFFFF),
	    locPhysAddr);
	HDEBUG(KERN_ALERT HERE "SGL: col=0x%X, row=0x%X, bank=0x%X, "
	    "rsrow=0x%X, dimm=0x%X\n",
	    Sgl.s.col, Sgl.s.row, Sgl.s.bank, Sgl.s.rsrows, Sgl.s.dimm);
	dimm = (int)Sgl.s.dimm;
	n_cols = unitDet->Dimm[dimm].cols_total;
	bytes_available = (n_cols << 3) - (int)(Sgl.s.col << 3);
	bytes_needed = (int)size;
	if (bytes_needed > bytes_available)
	{
	    HDEBUG(KERN_ALERT HERE "Not enough space in row! need %d,"
		" avail %d\n", bytes_needed, bytes_available);
	    bytes_to_send = (u64)bytes_available;
	    size -= bytes_available;
	}
	else
	{
	    bytes_to_send = size;
	    finished = 1;
	}
	Sgl.s.interrupt = 0;

	/*
	 * set the pci address
	 */
	physicalAddress = virt_to_bus(buffer);
	writel(physicalAddress, pRTR + REG_SGL_LOW);
	writel(0, pRTR + REG_SGL_HIGH);
	HDEBUG(RW_DEBUG HERE "PCI = %08lX %08lX\n",
	    (u_long) (physicalAddress >> 32),
	    (u_long) (physicalAddress & 0xFFFFFFFF));
	buffer = buffer + bytes_to_send;
	locLinAddr = locLinAddr + bytes_to_send;

	/*
	 * set the local bus address and other flags
	 */
	if (finished)
	    Sgl.s.interrupt = 1;
	Sgl.s.read = rdwt;
	Sgl.s.count = bytes_to_send >> 3;
	HDEBUG(KERN_ALERT HERE "SGL: intr=%d, read=%d, count=%d (0x%X), "
	    "nios=%d, drive=%d, dimm=%d, rsrow=%d, bank=%d, row=0x%X, "
	    "col=0x%X\n",
	    Sgl.s.interrupt, Sgl.s.read, Sgl.s.count, (Sgl.s.count << 3),
	    Sgl.s.nios, Sgl.s.drive, Sgl.s.dimm, Sgl.s.rsrows, Sgl.s.bank,
	    Sgl.s.row, Sgl.s.col);

	writel(Sgl.qw, pRTR + REG_SGL_LOW);
	HDEBUG(RW_DEBUG HERE "SGL_LOW = %08X\n", (u32) (Sgl.qw & 0xFFFFFFFF));
	writel(Sgl.qw >> 32, pRTR + REG_SGL_HIGH);
	HDEBUG(RW_DEBUG HERE "SGL_HIGH = %08X\n", (u32) (Sgl.qw >> 32));

	/*
	 * show the contents of the dma register
	 */
	dma = readl(pRTR + REG_DMA);
	HDEBUG(RW_DEBUG HERE "dma = %s (0x%08lX)\n", dma_decode(dma), dma);
    }

    /*
     * start the xfer
     */
    HDEBUG(KERN_ALERT HERE "about to GO\n");
    writel(DMA_GO, pRTR + REG_DMA);
    return (0);
}


int
qikdata_handle_read(kdev_t i_rdev, u_char * buffer, int sector,
    int nr_sectors)
{

    HDEBUG(KERN_ALERT HERE "qikdata_handle_read Sector %d, count %d\n",
	sector, nr_sectors);

    return (Handle_DMA(QD_READ, i_rdev, buffer, sector, nr_sectors));
}


int
qikdata_handle_write(kdev_t i_rdev, u_char * buffer, int sector, int nr_sectors)
{
    HDEBUG(KERN_ALERT HERE "qikdata_handle_write Sector %d, count %d\n", sector,
	nr_sectors);

    return (Handle_DMA(QD_WRITE, i_rdev, buffer, sector, nr_sectors));
}
