/*
 * drivers/linux2.4/plx9054.h
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#ifndef DRIVERS_LINUX2_4_PLX9054_H
#define DRIVERS_LINUX2_4_PLX9054_H

/*
 * PCI Configuration Registers
 */

#define REG_PCIISR                          0x06
#define PCI_STATUS_DATA_PARITY_DETECTED     (1<<8)
#define PCI_STATUS_SIGNALED_TARGET_ABORT    (1<<11)
#define PCI_STATUS_RECEIVED_TARGET_ABORT    (1<<12)
#define PCI_STATUS_RECEIVED_MASTER_ABORT    (1<<13)
#define PCI_STATUS_SIGNALED_SYSTEM_ERROR    (1<<14)
#define PCI_STATUS_DETECTED_PARITY_ERROR    (1<<15)

#define REG_PVPDAD                  0x4E
#define VPDADDRESS                  0x7FFF
#define VPDFLAG                     (1<<15)
#define VPDFLAGREAD                 (0<<15)
#define VPDFLAGWRITE                (VPDFLAG)

#define REG_PVPDATA                 0x50

/*
 * Local Registers
 */

#define REG_PROTAREA                        0x0E

#define REG_LAS0BA                          0x04
#define SPACE0ENABLE                        (1<<0)
#define REMAPPCIADDRESS_MASK                0xFFFFFFF0

#define REG_LBRD0                           0x18
#define LOCALBUSWIDTH_8                     0x00
#define LOCALBUSWIDTH_16                    0x01
#define LOCALBUSWIDTH_32                    0x03
#define MEMORYSPACE0TA_READY_INPUTENABLE    (1<<6)
#define MEMORYSPACE0BTERM_INPUTENABLE       (1<<7)

#define REG_LAS1BA                          0xF4
#define SPACE1ENABLE                        (1<<0)

#define REG_LBRD1                           0xF8
#define MEMORYSPACE1TA_READY_INPUTENABLE    (1<<6)
#define MEMORYSPACE1BTERM_INPUTENABLE       (1<<7)

/*
 * Runtime Registers
 */

#define REG_MBOX0                           0x40

#define REG_MBOX1                           0x44

#define REG_INTCSR                          0x68
#define POWERMANAGEMENTINTERRUPT            (1<<5)
#define PCIINTERRUPTENABLE                  (1<<8)
#define PCIABORTINTERRUPTENABLE             (1<<10)
#define RETRYABORTENABLE                    (1<<12)
#define PCIDOORBELLINTERRUPTACTIVE          (1<<13)
#define LOCALINTERRUPTINPUTENABLE           (1<<11)
#define PCIABORTINTERRUPTACTIVE             (1<<14)
#define LOCALINTERRUPTOUTPUTENABLE          (1<<16)
#define LOCALDMACHANNEL0INTERRUPTENABLE     (1<<18)
#define DMACHANNEL0INTERRUPTACTIVE          (1<<21)
#define DMACHANNEL1INTERRUPTACTIVE          (1<<22)
#define DMA0BUSMASTERDURINGABORT            (1<<25)
#define DMA1BUSMASTERDURINGABORT            (1<<26)

#define REG_CNTRL                   0x6C
#define GENERALPURPOSEOUTPUT_0      0
#define GENERALPURPOSEOUTPUT_1      (1<<16)
#define GENERALPURPOSEINPUT         (1<<17)
#define USERiINPUTENABLE            (1<<18)
#define USERoOUTPUTENABLE           (1<<19)
#define SERIALEEPROMCLOCK           (1<<24)
#define SERIALEEPROMCHIPSELECT      (1<<25)
#define PCIADAPTERSOFTWARERESET     (1<<30)

/*
 * DMA Registers
 */

#define REG_DMAMODE0                0x80
/*      LOCALBUSWIDTH_32            (see LBRD0) */
#define TA_READY_INPUTENABLE        (1<<6)
#define BTERM_INPUTENABLE           (1<<7)
#define LOCALBURSTENABLE            (1<<8)
#define SCATTERGATHERMODE           (1<<9)
#define DONEINTERRUPTENABLE         (1<<10)
#define DEMANDMODE                  (1<<12)
#define DMACHANNEL0INTERRUPTSELECT  (1<<17)

#define REG_DMAPADR0                0x84
#define REG_DMALADR0                0x88
#define REG_DMASIZ0                 0x8C

#define REG_DMADPR0                 0x90
#define DESCRIPTORLOCATION_PCI      (1<<0)
#define DESCRIPTORLOCATION_LB       0L
#define ENDOFCHAIN                  (1<<1)
#define INTERRUPTAFTERTERMINLCOUNT  (1<<2)
#define DIRECTIONOFTRANSFER_READ    (1<<3)
#define DIRECTIONOFTRANSFER_WRITE   0L

#define REG_DMACSR0                 0xA8
#define CHANNEL0ENABLE              (1<<0)
#define CHANNEL0START               (1<<1)
#define CHANNEL0ABORT               (1<<2)
#define CHANNEL0CLEARINTERRUPT      (1<<3)
#define CHANNEL0DONE                (1<<4)

#endif /* DRIVERS_LINUX2_4_PLX9054_H */
