/*
 * programs/qd_types/main.c
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


int
main(int argc, char **argv)
{
    FILE *fp = stdout;

    switch (argc)
    {
    case 1:
	break;

    case 2:
	fp = fopen(argv[1], "w");
	if (!fp)
	{
	    fprintf(stderr, "open %s: %s", argv[1], strerror(errno));
	    exit(1);
	}
	break;

    default:
    	fprintf(stderr, "Usage: %s [ <filename> ]\n", argv[0]);
	exit(1);
    }

    /*
     * DO NOT generate the date into this file.  It messes up the
     * fingerprints and induces lots of unnecessary builds.
     */
    fprintf(fp,
"/*\n"
" * Warning: this file is generated.  Do not edit this file directly.\n"
" * If you find a prblem with this file, edit the programs/qd_types/main.c\n"
" * file.  You will need to be VERY careful, because it must work across\n"
" * a wide range of operating systems and CPU types.\n"
" *\n"
" * Warning: I told you not to edit this file, didn't I?\n"
" *\n"
" * I thought about adding the directory name to the #define, below, but\n"
" * then I realised that I only want *one* of the files per architecture\n"
" * for the whole project, so I left it out.\n"
" */\n");

    fprintf(fp, "#ifndef QD_TYPE_H\n");
    fprintf(fp, "#define QD_TYPE_H\n");
    fprintf(fp, "\n");

    fprintf(fp, "typedef unsigned char          qd_uint8_t;\n");
    fprintf(fp, "typedef   signed char          qd_int8_t;\n");
    if (sizeof(int) == 2)
    {
	fprintf(fp, "typedef unsigned           int qd_uint16_t;\n");
	fprintf(fp, "typedef                    int qd_int16_t;\n");
    }
    else
    {
	fprintf(fp, "typedef unsigned short     int qd_uint16_t;\n");
	fprintf(fp, "typedef          short     int qd_int16_t;\n");
    }
    if (sizeof(int) == 4)
    {
	fprintf(fp, "typedef unsigned           int qd_uint32_t;\n");
	fprintf(fp, "typedef                    int qd_int32_t;\n");
    }
    else
    {
	fprintf(fp, "typedef unsigned long      int qd_uint32_t;\n");
	fprintf(fp, "typedef          long      int qd_int32_t;\n");
    }
    if (sizeof(long int) == 8)
    {
	fprintf(fp, "typedef unsigned long      int qd_uint64_t;\n");
	fprintf(fp, "typedef          long      int qd_int64_t;\n");
    }
    else
    {
	fprintf(fp, "typedef unsigned long long int qd_uint64_t;\n");
	fprintf(fp, "typedef          long long int qd_int64_t;\n");
    }
    fprintf(fp, "\n");
    fprintf(fp, "#endif /* QD_TYPE_H */\n");

    if (fflush(fp))
    {
	fprintf(stderr, "write %s: %s", argv[1], strerror(errno));
	exit(1);
    }
    if (fp != stdout && fclose(fp))
    {
	fprintf(stderr, "close %s: %s", argv[1], strerror(errno));
	exit(1);
    }
    exit(0);
    return 0;
}
