/*
 * drivers/linux2.4/qikdata_hba.h
 *
 *	Copyright (C) 2003 Platypus Technology Holdings, Inc.
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer
 *	in the documentation and/or other materials provided with the
 *	distribution.
 *
 *	Neither the name of the Platypus Technology Holdings, Inc. nor
 *	the names of its contributors may be used to endorse or promote
 *	products derived from this software without specific prior
 *	written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *	CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 *	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * If you find a bug in this device driver, please contact support@platypus.net
 * quoting the version number (1.1.D585).  If you fix a bug in this device
 * driver, please send patches to the same address.
 */

#ifndef DRIVERS_LINUX2_4_QIKDATA_HBA_H
#define DRIVERS_LINUX2_4_QIKDATA_HBA_H

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Set all Build options here
 */

/*#define _QIK64_EMPTY_REQUEST */
#define _USE_DMA_READ
#define _USE_DMA_WRITE
#define DMA_POLL_DELAY_5U  (200*1000)

#define _ALWAYS_COLD_BOOT
#define _NEVER_ENABLE_ECC
#ifdef _NEVER_ENABLE_ECC
#define _NO_SDRAM_INITIALISATION
#define _FORMAT_FAT16
#endif

/*
 * BAR Definitions
 */

#define LASmem                      0
#define RTRmem                      1

#define LAS                     LASmem
#define RTR                     RTRmem

/*
 *      EEPROM definitions
 */

#define VENDOR_PLATYPUS		0x118b
#define QIK64_TYPE_2_SLOT       0x0302
#define QIK64_TYPE_2_SLOT       0x0302
#define QIKDATA_HBA_CARD        0x0401

#define DRIVE_MAGIC 		0x59414251

/*
 *      EVAL REGISTER defines
 *      - all register offsets are in (32-bit) dwords
 */

#define REG_ID                  (0x00)
#define ID_ID_MASK              0x0000FF00
#define ID_VERSION_MASK         0x000000FF

#define REG_IO                  (0x10)
#define IO_LINK_RESET		(1<<25)
#define IO_SLOT32               (1<<20)
#define IO_TDI                  (1<<19)
#define IO_EXT_12V_FAIL         (1<<18)
#define IO_M66_ENA              (1<<17)
#define IO_I2C_SDA_IN           (1<<16)
#define IO_WARM_BOOT            (1<<15)
#define IO_LED_OVERRIDE         (1<<14)
#define IO_LB_CLOCK_SPEED_1     (1<<13)
#define IO_LB_CLOCK_SPEED_0     (1<<12)
#define IO_TMS                  (1<<11)
#define IO_TCK                  (1<<10)
#define IO_TDO                  (1<<9)
#define IO_DAT_ECC              (1<<8)
#define IO_CMD_ECC              (1<<7)
#define IO_BUZZER               (1<<6)
#define IO_LED_READ             (1<<5)
#define IO_LED_WRITE            (1<<4)
#define IO_LED_ECC              (1<<3)
#define IO_LED_STANDBY          (1<<2)
#define IO_LED_GOOD             (1<<1)
#define IO_nLED_ATTN            (1<<0)

#define REG_DMA                 (0x20)
#define DMA_SGL_INTERRUPT       (1<<31)
#define DMA_GO                  (1<<30)
#define DMA_FIFO_CLEAR          (1<<29)
#define DMA_ABORT_INTERRUPT     (1<<28)
#define DMA_PCI_CORE_INTERRUPT  (1<<27)
#define DMA_ECC_INTERRUPT       (1<<26)
#define DMA_UNCORRECTABLE       (1<<25)
#define DMA_CORRECTABLE         (1<<24)
#define DMA_INTERRUPT_PENDING   (1<<16)
#define DMA_SGL_HIGH_USED_MASK  0x0003FE00    /* 17:9 */
#define DMA_SGL_LOW_USED_MASK   0x000001FF    /* 8:0 */

#define REG_PAGE                (0x30)
#define PAGE_MASK_NIOS          0x08000000    /* 27 */
#define PAGE_MASK_DRIVE         0x07e00000    /* 26:21 */
#define PAGE_MASK_DIMM          0x001e0000    /* 20:17 */
#define PAGE_MASK_RSROW         0x00010000    /* 16 */
#define PAGE_MASK_BANK          0x0000c000    /* 15:14 */
#define PAGE_MASK_ROW           0x00003fff    /* 13:0 */
/* when accessing the NIOS space, the */
/* following bits have different meanings */
#define PAGE_MASK_EXTENDED      0x00010000    /* 16 */
#define PAGE_MASK_DIMM_SPACE    0x00008000    /* 15 */

#define REG_SGL_LOW             (0x40)
#define REG_SGL_HIGH            (0x50)

#define REG_STATUS_LOW          (0x50)
#define REG_STATUS_HIGH         (0x60)

#define STATLOW_MASK_FATAL      (1<<1)
#define STATLOW_MASK_NONFATAL   (1<<0)

#define REG_SDRAM               (0x60)
#define SDRAM_MASK              0x00000FFF    /* 11:0 */

#define REG_ECC                 (0x70)
#define ECC_MASK_DATA_UNCORR	0xff000000    /* 31:24 */
#define ECC_SHIFT_DATA_UNCORR	24
#define ECC_MASK_DATA_CORR      0x00ff0000    /* 23:16 */
#define ECC_SHIFT_DATA_CORR     16
#define ECC_MASK_CMD_UNCORR     0x0000ff00    /* 15:8 */
#define ECC_SHIFT_CMD_UNCORR    8
#define ECC_MASK_CMD_CORR       0x000000ff    /* 7:0 */
#define ECC_SHIFT_CMD_CORR      0

typedef union PAGE_REG PAGE_REG;
union PAGE_REG
{
    u32 dw;

    struct
    {
        u32     row:14;
        u32     bank:2;
        u32     rsrow:1;
        u32     dimm:4;
	u32     drive:6;
	u32     nios:1;
        u32     _unused:4;
    }       s;

};
typedef PAGE_REG *PPAGE_REG;

typedef union SGL_ENTRY SGL_ENTRY;
union SGL_ENTRY
{
    u64 qw;

    struct
    {
	u64     _reserved:3;
	u64     col:13;
	u64     row:14;
	u64     bank:2;
	u64     rsrows:1;
	u64     dimm:4;
	u64     drive:6;
	u64     nios:1;
	u64     count:12;
	u64     read:1;
	u64     interrupt:1;
	u64     _unused:6;
    }       s;
};
typedef SGL_ENTRY *PSGL_ENTRY;

#ifdef __cplusplus
}
#endif

#endif /* DRIVERS_LINUX2_4_QIKDATA_HBA_H */
