/* $Id: lkst.c,v 1.10 2002/06/28 11:27:55 indoh Exp $ */ 
/* linux/drivers/lkst/lkst.c */

/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2001-2002 ALL RIGHTS RESERVED.
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by K.Serizawa <serizawa@sdl.hitachi.co.jp>
 *
 ************************************************************/

#include <linux/init.h>
#include <linux/sched.h>
#include <linux/sysrq.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/lkst_private.h>
#include <linux/stat.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/spinlock.h>
#include <linux/utsname.h>

// TODO: check architecture independency
#include <asm/uaccess.h>
#include <asm/atomic.h>

				/* Array of mnemonics of event types */
#define LKST_ETYPE_DEF_MNEMONICS

#include <linux/lkst_events.h>

				/* Array of descriptions of event types */
#define LKST_ETYPE_DEF_DESCRIPTIONS

#include <linux/lkst_events.h>

		/* Error code for lkst_evhandlerprim_buffer_shift_inline() */
#define LKST_BUFFER_SHIFT_ERRNO_SUCCESS	0
#define LKST_BUFFER_SHIFT_ERRNO_ECPU	(-1)
#define LKST_BUFFER_SHIFT_ERRNO_ECBUF	(-2)
#define LKST_BUFFER_SHIFT_ERRNO_ENBUF	(-3)
#define LKST_BUFFER_SHIFT_ERRNO_EBUFC	(-4)

		/* Error code for lkst_evhandlerprim_maskset_xchg_inline() */
#define LKST_MASKSET_XCHG_ERRNO_SUCCESS	0
#define LKST_MASKSET_XCHG_ERRNO_WCID	(-1)
#define LKST_MASKSET_XCHG_ERRNO_WCPT	(-2)
#define LKST_MASKSET_XCHG_ERRNO_ECPT	(-3)
#define LKST_MASKSET_XCHG_ERRNO_ENID	(-4)
#define LKST_MASKSET_XCHG_ERRNO_ENPT	(-5)
#define LKST_MASKSET_XCHG_ERRNO_ESTAT	(-6)


#ifdef CONFIG_DEBUG_KERNEL_STATE_TRACE
extern struct lkst_header lkst;
#else /* !CONFIG_DEBUG_KERNEL_STATE_TRACE */
#define lkst (*((struct lkst_header *)get_lkst()))
#endif /* CONFIG_DEBUG_KERNEL_STATE_TRACE */

#define cpu_khz (lkst_get_cpu_khz())

EXPORT_SYMBOL(lkst_trc_status);
EXPORT_SYMBOL(lkst_trc_start);
EXPORT_SYMBOL(lkst_trc_stop);
EXPORT_SYMBOL(lkst_maskset_xchg);
EXPORT_SYMBOL(lkst_maskset_set);
EXPORT_SYMBOL(lkst_maskset_read);
EXPORT_SYMBOL(lkst_maskset_write);
EXPORT_SYMBOL(lkst_maskset_delete);
EXPORT_SYMBOL(lkst_maskset_list);
EXPORT_SYMBOL(lkst_evhandler_ctrl);
EXPORT_SYMBOL(lkst_evhandler_list);
EXPORT_SYMBOL(lkst_evhandler_register);
EXPORT_SYMBOL(lkst_evhandler_unregister);
EXPORT_SYMBOL(lkst_evhandler_get_id);
EXPORT_SYMBOL(lkst_buffer_read);
EXPORT_SYMBOL(lkst_buffer_create);
EXPORT_SYMBOL(lkst_buffer_shift);
EXPORT_SYMBOL(lkst_buffer_delete);
#if DEBUG
EXPORT_SYMBOL(lkst_buffer_seek);
#endif /* DEBUG */
EXPORT_SYMBOL(lkst_buffer_list);
EXPORT_SYMBOL(lkst_buffer_dump_current_txt);
EXPORT_SYMBOL(lkst_evhandlerprim_select_buffer_cpu);
EXPORT_SYMBOL(lkst_evhandlerprim_select_buffer_id);
EXPORT_SYMBOL(lkst_evhandlerprim_select_buffer);
EXPORT_SYMBOL(lkst_evhandlerprim_recid_inc);
EXPORT_SYMBOL(lkst_evhandlerprim_get_write_entry);
EXPORT_SYMBOL(lkst_evhandlerprim_getw_buffer);
EXPORT_SYMBOL(lkst_evhandlerprim_getw_buffer_cpu);
EXPORT_SYMBOL(lkst_evhandlerprim_putw_buffer);
EXPORT_SYMBOL(lkst_evhandlerprim_daemon_wakeup);
EXPORT_SYMBOL(lkst_evhandlerprim_entry_log_l2);
EXPORT_SYMBOL(lkst_evhandler_default);
EXPORT_SYMBOL(lkst_evhandler_default_ctrl);
EXPORT_SYMBOL(lkst_evhandler_buffer_shift);
EXPORT_SYMBOL(lkst_evhandler_maskset_xchg);
EXPORT_SYMBOL(lkst_evhandler_buffer_shift_dw);
EXPORT_SYMBOL(lkst_event_type_mnemonics);
EXPORT_SYMBOL(lkst_event_type_descriptions);
EXPORT_SYMBOL(lkst_hook_init_module);
EXPORT_SYMBOL(lkst_hook_reset_all);
#if DEBUG
EXPORT_SYMBOL(lkst_evhandler_exception);
#endif /* DEBUG */

int lkst_init_flag = LKST_FLAG_NO; // TODO: not used
static size_t lkst_init_buffer_size = LKST_BUFFER_SIZE_DEFAULT;
static struct lkst_event_buffer lkst_stage0_event_buffer = {
	writep:{LKST_BUFFER_HEADSPACE},
	size:LKST_BUFFER_SIZE_DEFAULT,
	readp:LKST_BUFFER_HEADSPACE,
	id:0,next:LKST_BUFFER_ID_VOID,prev:LKST_BUFFER_ID_VOID,
	eyecatcher:{'L','K','B','F'},
	cpu:0,
	last_recid:0,
	last_readp:(LKST_BUFFER_HEADSPACE - sizeof (struct lkst_event_record)) & (LKST_BUFFER_SIZE_DEFAULT - 1),
	xtime:{0, 0},
	cmos_time_flag:1,
	buf_ver:LKST_BUF_VER,
	rcnt:{0},
	wcnt:{0}
};
spinlock_t lkst_buffer_lock = SPIN_LOCK_UNLOCKED;
spinlock_t lkst_status_lock = SPIN_LOCK_UNLOCKED;

/* definitions of special masksets. */
/*
 * maskset_rnothing (id: 0(LKST_MASKSET_ID_RNOTHING))
 * a maskset to record no events.
 */
static const lkst_evhandler_id maskset_rnothing[] = {
	[0 ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_VOID
	    /* caution: LKST_ETYPE_MAX is significant
	     *   to decide the size of this maskset.
	     */
};

/*
 * maskset_rall (id: 1(LKST_MASKSET_ID_RALL))
 * a maskset to record all events.
 */
static const lkst_evhandler_id maskset_rall[] = {
	[0 ... LKST_ETYPE_LKST_BUFF_OVFLOW - 1] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_LKST_BUFF_OVFLOW] = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_LKST_BUFF_OVFLOW + 1 ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_DEFAULT
	    /* caution: LKST_ETYPE_MAX is significant, ditto. */
};

/*
 * maskset_rdefault (id: 2(LKST_MASKSET_ID_RDEFAULT))
 * a maskset selected initially.
 */
static const lkst_evhandler_id maskset_rdefault[] = {
	/* Default Maskset omits lock operation events */

	[0 ... LKST_ETYPE_MEM_DO_NOPAGE-1 ] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_MEM_DO_NOPAGE ... LKST_ETYPE_NET_PKTSENDI-1 ] = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_NET_PKTSENDI] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_NET_PKTRECV]  = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_NET_PKTRECV+1 ... LKST_ETYPE_LK_SPINLOCK -1] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_LK_SPINLOCK ... LKST_ETYPE_O_PORTOUT] = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_O_PORTOUT + 1 ... LKST_ETYPE_TIMER_ADD - 1 ] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_TIMER_ADD ... LKST_ETYPE_LKST_INIT - 1 ] = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_LKST_INIT ... LKST_ETYPE_LKST_BUFF_OVFLOW - 1] = LKST_EVHANDLER_ID_DEFAULT,
	[LKST_ETYPE_LKST_BUFF_OVFLOW] = LKST_EVHANDLER_ID_VOID,
	[LKST_ETYPE_LKST_BUFF_OVFLOW + 1 ... LKST_ETYPE_MAX] = LKST_EVHANDLER_ID_DEFAULT
	    /* caution: LKST_ETYPE_MAX is significant, ditto. */
};

/* LKST device file_operations */
static struct file_operations lkst_fops = {
	ioctl:lkst_ioctl,
	open:lkst_open,
	read:lkst_read,
	write:NULL,
	poll:NULL,
	release:lkst_close,
	lock:NULL,
};

#if DEBUG
/* for ioctl(LKST_IOC_DIAG) */
#define LKST_IOC_DIAG		0xfe

struct lkst_diag_arg {
	int (*func) (char *arg);
	char arg[128];
} lkst_diag_argdata[LKST_CPU_MAX];

static int lkst_ioc_diag(long arg);
#endif /* DEBUG */

#ifdef CONFIG_MAGIC_SYSRQ
static void sysrq_handle_lkst_dump_buffer(int key, struct pt_regs *pt_regs,
					  struct kbd_struct *kbd,
					  struct tty_struct *tty)
{
	lkst_buffer_dump_current_txt();
}

static struct sysrq_key_op sysrq_lkst_dump_buffer_op = {
	handler:sysrq_handle_lkst_dump_buffer,
	help_msg:"lkstView",
	action_msg:"Dump Buffer\n",
};
#endif /* CONFIG_MAGIC_SYSRQ */

/* for devfs support */
#ifdef CONFIG_DEVFS_FS
static devfs_handle_t lkst_devfs_handle = NULL;
#endif	

/*********************************************************************
 *  Kernel hook functions
 *********************************************************************/


/* 
 *
 *  GKHI hook function
 *
 */

extern void lkst_trc_hook_init(void);
extern void lkst_trc_hook_cleanup(void);

/* Exchange of hook exit */
static __inline__ int lkst_hook_exit_xchg(lkst_event_type etype, 
					  lkst_evhandler_id newid,
					  lkst_evhandler_id oldid)
{
	int ret = 0;
	hook_rec_t* phookrec = lkst_hook_info_table[etype].phookrec;
	lkst_evhandler_func new_evhandler;

	if(phookrec == NULL) return 0;

	/* unregister old exit */
	if(oldid <= LKST_EVHANDLER_ID_MAX){
		hook_exit_deregister(phookrec);
		memset(phookrec, 0, sizeof(hook_rec_t));
	}

	/* register new exit */
	if(newid <= LKST_EVHANDLER_ID_MAX){
		if (!LKST_POINTER_VERIFY
		    (lkst.evhandler_table[newid].evhandler, new_evhandler)) {
			/* Handler(ID=newid) does not exist. */
			return 0;
		}
		phookrec->hook_exit = new_evhandler;
		/* phookrec->hook_exit_name = lkst.evhandler_table[newid].name; */
		phookrec->hook_link = NULL;
		ret = _hook_exit_register(lkst_hook_info_table[etype].hook_id,
					  phookrec, 0);
		ret += hook_exit_arm(phookrec);
	}
	return ret;
}


/* 
 * Exchange maskset
 *  Called from lkst_trc_init_stage0, lkst_evhandlerprim_maskset_xchg_inline,
 *  lkst_trc_start_core and lkst_trc_stop_core.
 */
int lkst_hook_maskset_xchg(lkst_maskset_id mask_id_new,
			   lkst_maskset_id mask_id_old)
{
	int i, ret = 0;
	lkst_maskset *maskset_old;
	lkst_maskset *maskset_new;
	lkst_evhandler_id evh_id_old;
	lkst_evhandler_id evh_id_new;

	if(mask_id_old == mask_id_new) return 0;

	if (mask_id_old > LKST_MASKSET_ID_MAX) {
		mask_id_old = LKST_MASKSET_ID_RNOTHING;
	}
	if (mask_id_new > LKST_MASKSET_ID_MAX) {
		mask_id_new = LKST_MASKSET_ID_RDEFAULT;
	}

	maskset_old = lkst.maskset_table[mask_id_old].masksetp;
	maskset_new = lkst.maskset_table[mask_id_new].masksetp;

	for (i = 0; i < LKST_MASKSET_TABLE_LEN_MAX; i++) {
		evh_id_old = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset_old, i);
		evh_id_new = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset_new, i);

		if(evh_id_old != evh_id_new) {
			ret = lkst_hook_exit_xchg(i, evh_id_new, evh_id_old);
		}
	}

	if (ret) {
		printk(KERN_ERR "LKST error(hook_maskset_xchg): hook_maskset_xchg failed: %u\n", ret);
		return -EINVAL;
	}
	return 0;
}

/* 
 * Update hook exit
 *  Called from lkst_evhandler_register
 */
static __inline__ int lkst_hook_exit_update(lkst_evhandler_id evh_id)
{
	lkst_evhandler_id evhandler_id;
	lkst_maskset *maskset = NULL;
	int i, ret = 0;

	if (!LKST_POINTER_VERIFY(lkst.current_masksetp, maskset)) {
		/* lkst_evhandler_register is called from stage0 */
		return 0;
	}

	for(i=0; i<LKST_MASKSET_TABLE_LEN_MAX; i++){
		evhandler_id = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, i);
		if(evhandler_id == evh_id){
			ret += lkst_hook_exit_xchg(i, evh_id, 0);
		}
	}

	if (ret) {
		printk(KERN_ERR "LKST error(hook_exit_update): hook_exit_update failed: %u\n", ret);
		return -EINVAL;
	}
	return 0;
}

/* 
 * Initialize hook in the modules
 */
void lkst_hook_init_module(lkst_event_type etype)
{
	lkst_evhandler_id evhandler_id;
	lkst_maskset *maskset = NULL;

	if ((etype > LKST_ETYPE_MAX) || (etype < 0)) {
		printk(KERN_ERR "LKST error(lkst_hook_init_module): invalid event id %u\n", etype);
		return;
	}

	if (!LKST_POINTER_VERIFY(lkst.current_masksetp, maskset)) {
		return;
	}

	evhandler_id = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, etype);
	lkst_hook_exit_xchg(etype, evhandler_id, LKST_EVHANDLER_ID_MAX + 1);
}

/* 
 * Reset all hook
 */
void lkst_hook_reset_all(void)
{
	lkst_hook_maskset_xchg(LKST_MASKSET_ID_RNOTHING, LKST_MASKSET_ID_RALL);
	lkst_hook_maskset_xchg(lkst.current_maskset_id, LKST_MASKSET_ID_RNOTHING);
}


/*********************************************************************
 * evhandler primitive functions
 *********************************************************************/

/*
 * function lkst_evhandlerprim_recid_inc
 * A primitive part of an Event-handler. This returns ID of the newest 
 * event record.
 * 
 *  Arguments
 *    NONE
 *
 *  returns
 *    recid:		ID of the newest event record.
 * 
 */
static __inline__ int lkst_evhandlerprim_recid_inc_inline(void)
{
	return (lkst.cpu_recid[LKST_CPU_RECID_SIZE*smp_processor_id()]++);
}
int lkst_evhandlerprim_recid_inc()
{
	return (lkst_evhandlerprim_recid_inc_inline());
}

/*
 * function lkst_evhandlerprim_select_buffer_cpu
 * A primitive part of an Event-handler. This returns address of a buffer 
 * to write on a specified CPU.
 * 
 *  Arguments
 *    cpu {IN}:	num of CPU
 *
 *  returns
 *    bufferp: address of a buffer to write.
 * 
 */
static __inline__ struct lkst_event_buffer
*lkst_evhandlerprim_select_buffer_cpu_inline(int cpu)
{
	lkst_buffer_id buffer_id;
	struct lkst_event_buffer *bufferp;

	buffer_id =lkst.current_buffer_id[cpu];

	if (buffer_id > LKST_BUFFER_ID_MAX)
		buffer_id = 0;  /* buffer_id:0 is static, always exists.*/
	if (!LKST_POINTER_VERIFY(lkst.buffer_table[buffer_id].bufferp, bufferp))
		bufferp = lkst.buffer_table[0].bufferp;

	return (bufferp);
}

struct lkst_event_buffer
*lkst_evhandlerprim_select_buffer_cpu(int cpu)
{
	return (lkst_evhandlerprim_select_buffer_cpu_inline(cpu));
}


/*
 * function lkst_evhandlerprim_select_buffer
 * A primitive part of an Event-handler. This returns address of a buffer 
 * to write.
 * 
 *  Arguments
 *    NONE
 *
 *  returns
 *    buffer: address of a buffer to write.
 * 
 */
static __inline__ struct lkst_event_buffer
*lkst_evhandlerprim_select_buffer_inline(void)
{
	return (lkst_evhandlerprim_select_buffer_cpu_inline(smp_processor_id()));
}
struct lkst_event_buffer
*lkst_evhandlerprim_select_buffer(void)
{
	return (lkst_evhandlerprim_select_buffer_inline());
}


/*
 * function lkst_evhandlerprim_select_buffer_id
 * A primitive part of an Event-handler. This returns address of specified 
 * buffer to write.
 * 
 *  Arguments
 *    id {IN}:	buffer id.
 *
 *  returns
 *    bufferp: address of a buffer to write.
 * 
 */
static __inline__ struct lkst_event_buffer
*lkst_evhandlerprim_select_buffer_id_inline(lkst_buffer_id id)
{
	struct lkst_event_buffer *bufferp;

	if (id > LKST_BUFFER_ID_MAX)
		id = 0;  /* id:0 is static, always exists.*/
	if (!LKST_POINTER_VERIFY(lkst.buffer_table[id].bufferp, bufferp))
		bufferp = lkst.buffer_table[0].bufferp;

	return (bufferp);
}
struct lkst_event_buffer
*lkst_evhandlerprim_select_buffer_id(lkst_buffer_id id)
{
	return (lkst_evhandlerprim_select_buffer_id_inline(id));
}


/*
 * function lkst_evhandlerprim_get_write_entry_inline
 * A primitive part of an Event-handler. This check current_buffer
 * and increase write-pointer of event buffer and return address of
 * entry of event buffer to write. 
 * 
 *  Arguments
 *    event_type:	event type
 *
 *  returns
 *    event_entry_p:	address of entry to write.
 *      NULL		found no entry to write.
 */
static __inline__ struct lkst_event_record
*lkst_evhandlerprim_get_write_entry_inline(int event_type)
{
	struct lkst_event_buffer *buffer;
	struct lkst_event_record *event_entry_p;

	buffer = lkst_evhandlerprim_select_buffer_inline();
	
	event_entry_p =
		lkst_evhandlerprim_entry_next(buffer, event_type);

	return (event_entry_p);
}
struct lkst_event_record
*lkst_evhandlerprim_get_write_entry(int event_type)
{
	return (lkst_evhandlerprim_get_write_entry_inline(event_type));
}

/*
 * function lkst_evhandlerprim_getw_buffer_cpu_inline
 * A primitive part of an Event-handler. This marks for buffer writing,
 * and returns address of a buffer to write on a specified CPU.
 * 
 *  Arguments
 *    cpu {IN}:	num of CPU
 *
 *  returns
 *    buffer: address of a buffer to write.
 *
 */
static __inline__ struct lkst_event_buffer
*lkst_evhandlerprim_getw_buffer_cpu_inline(int cpu)
{
	struct lkst_event_buffer *buffer;

	buffer = lkst_evhandlerprim_select_buffer_cpu_inline(cpu);

	return buffer;
}
struct lkst_event_buffer *lkst_evhandlerprim_getw_buffer_cpu(int cpu)
{
	return (lkst_evhandlerprim_getw_buffer_cpu_inline(cpu));
}

/*
 * function lkst_evhandlerprim_getw_buffer_inline
 * A primitive part of an Event-handler. This marks for buffer writing,
 * and returns address of a buffer to write on a CPU correspond to the event.
 * This function is obsoleted because racing between buffer deletion and buffer
 * writing is resolved by using "lkst.refcnt".
 * 
 *  Arguments
 *    NONE
 *
 *  returns
 *    buffer: address of a buffer to write.
 *
 */
static __inline__ struct lkst_event_buffer
*lkst_evhandlerprim_getw_buffer_inline(void)
{
	struct lkst_event_buffer *buffer;

	buffer = lkst_evhandlerprim_select_buffer_cpu_inline(smp_processor_id());

	return buffer;
}
struct lkst_event_buffer *lkst_evhandlerprim_getw_buffer(void)
{
	return (lkst_evhandlerprim_getw_buffer_inline());
}

/*
 * function lkst_evhandlerprim_putw_buffer_inline
 * A primitive part of an Event-handler. This removes a mark of buffer writing
 * This function is obsoleted because racing between buffer deletion and buffer
 * writing is resolved by using "lkst.refcnt".
 *
 *  Arguments
 *    buffer: address of a buffer which has a mark of writing.
 *
 *  returns
 *    NONE
 *
 */
static __inline__ void lkst_evhandlerprim_putw_buffer_inline(
		       struct lkst_event_buffer *buffer)
{
}
void lkst_evhandlerprim_putw_buffer(struct lkst_event_buffer *buffer)
{
	lkst_evhandlerprim_putw_buffer_inline(buffer);
}


/*
 * function lkst_evhandlerprim_entry_log_l2_inline
 * A primitive part of an Event-handler. This write an event log to 
 * a currently selected buffer correspond to a cpu on which the event
 * has occurred.
 * 
 *  Arguments
 *    event_type {IN}:	corresponding event type.
 *    arg1-4 {IN}:	arguments acquired at a trace point
 *
 *  returns
 *    NONE
 */
static __inline__ void
lkst_evhandlerprim_entry_log_l2_inline(int event_type,
				       lkst_arg_t arg1, lkst_arg_t arg2,
				       lkst_arg_t arg3, lkst_arg_t arg4)
{
	struct lkst_event_buffer *buffer;
	struct lkst_event_record *event_entry_p;
	unsigned long flags;
	__volatile__ int recid;

	LKST_CLI(flags);

	buffer = lkst_evhandlerprim_getw_buffer_inline();

	recid = lkst_evhandlerprim_recid_inc_inline();
	event_entry_p = lkst_evhandlerprim_entry_next(buffer, event_type);

	if (event_entry_p)
		lkst_evhandlerprim_entry_log_l1(event_entry_p, recid,
						event_type, arg1, arg2, arg3, arg4);

	lkst_evhandlerprim_putw_buffer_inline(buffer);

	LKST_STI(flags);
}
void lkst_evhandlerprim_entry_log_l2(int event_type,
				     lkst_arg_t arg1, lkst_arg_t arg2,
				     lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandlerprim_entry_log_l2_inline(event_type,
					       arg1, arg2, arg3, arg4);
}


/*
 * function lkst_evhandlerprim_buffer_shift_inline
 * A primitive part of an Event-handler. This primitive
 * shifts buffer in the chain of specified cpu.
 * This primitive is for LKST internal use.
 * 
 *  Arguments
 *    cpu {IN}:	cpu id the buffer chain bound.
 *
 *  returns
 *    ret: 	error code (private).
 */

static __inline__ int
lkst_evhandlerprim_buffer_shift_inline(int cpu)
{
	struct lkst_event_buffer *buffer, *next_buffer;
	lkst_buffer_id buffer_id;
	unsigned long flags;
	int	ret = LKST_BUFFER_SHIFT_ERRNO_SUCCESS;

	if (cpu >= smp_num_cpus || cpu < 0) {
		ret = LKST_BUFFER_SHIFT_ERRNO_ECPU;
		return ret;
	}

	LKST_CLI(flags);
	atomic_inc(&(lkst.refcnt));

	buffer_id = lkst.current_buffer_id[cpu];
	if (!((buffer_id <= LKST_BUFFER_ID_MAX) &&
	    (LKST_POINTER_VERIFY(lkst.buffer_table[buffer_id].bufferp, buffer)))) {

		ret = LKST_BUFFER_SHIFT_ERRNO_ECBUF;
		atomic_dec(&(lkst.refcnt));
		LKST_STI(flags);
		return ret;
	}

	buffer_id = buffer->next;
	if (!((buffer_id <= LKST_BUFFER_ID_MAX) &&
	    (LKST_POINTER_VERIFY(lkst.buffer_table[buffer_id].bufferp, next_buffer)))) {
		ret = LKST_BUFFER_SHIFT_ERRNO_ENBUF;
		atomic_dec(&(lkst.refcnt));
		LKST_STI(flags);
		return ret;
	}

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_BUFF_SHIFT,
		       LKST_ARG32(buffer->id, 0),
		       LKST_ARG32(next_buffer->id, 0),
		       LKST_ARG32(buffer, 0), LKST_ARG32(next_buffer,0));

	/* increment next buffer write count */
	atomic_inc(&(next_buffer->wcnt));

	/* move write pointer */
	atomic_set(&(next_buffer->writep), LKST_BUFFER_HEADSPACE);

	/* change current buffer */
	buffer_id = lkst.current_buffer_id[cpu] = next_buffer->id;

	/* increment previous buffer write count */
	atomic_inc(&(buffer->wcnt));

#if DEBUG
	/* Reconfirm this buffer is not been deleting */
	if (!LKST_POINTER_VERIFY(lkst.buffer_table[buffer_id].bufferp, next_buffer)) {
		ret = LKST_BUFFER_SHIFT_ERRNO_EBUFC;

			/* Restore current buffer id. */
		lkst.current_buffer_id[cpu] = buffer->id;

		atomic_dec(&(lkst.refcnt));
		LKST_STI(flags);
		return ret;
	}
#endif /* DEBUG */

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_BUFF_SHIFT,
		       LKST_ARG32(buffer->id, 0),
		       LKST_ARG32(next_buffer->id, 0),
		       LKST_ARG32(buffer, 0), LKST_ARG32(next_buffer,0));

	atomic_dec(&(lkst.refcnt));
	LKST_STI(flags);
	return (LKST_BUFFER_SHIFT_ERRNO_SUCCESS);
}


/*
 * function lkst_evhandlerprim_maskset_xchg_inline
 * A primitive part of an Event-handler. This primitive
 * change current selected maskset to a specified id.
 * This primitive is for LKST internal use.
 * 
 *  Arguments
 *    id {IN}:	An id of a maskset to be selected next.
 *
 *  returns
 *    ret: 	error code (private).
 */

static __inline__ int
lkst_evhandlerprim_maskset_xchg_inline(lkst_maskset_id * id, int xchg_hook)
{
	unsigned long flags;
	int	ret = LKST_MASKSET_XCHG_ERRNO_SUCCESS;
	lkst_maskset_id old_id = lkst.current_maskset_id, new_id = *id;
	volatile lkst_maskset *old_maskset = NULL, *new_maskset = NULL;

	LKST_CLI(flags);

	/* Verify old maskset pointer. */
	if (LKST_POINTER_VERIFY(lkst.current_masksetp, old_maskset)) {
		/* Verify old maskset id. */
		if (old_maskset != lkst.maskset_table[old_id].masksetp) {
			ret = LKST_MASKSET_XCHG_ERRNO_WCPT;
			old_id = LKST_MASKSET_ID_RDEFAULT;
		} else {
			if (old_id > LKST_MASKSET_ID_MAX) {
				ret = LKST_MASKSET_XCHG_ERRNO_WCID;
				old_id = LKST_MASKSET_ID_RDEFAULT;
			}
		}
	} else {
		ret = LKST_MASKSET_XCHG_ERRNO_WCPT;
		old_id = LKST_MASKSET_ID_RDEFAULT;
	}
	/* if old maskset has not been initialized, return as error. */
	if (!LKST_POINTER_VERIFY
	    (lkst.maskset_table[old_id].masksetp, old_maskset)) {
		ret = LKST_MASKSET_XCHG_ERRNO_ECPT;
		LKST_STI(flags);
		return ret;
	}

	if (new_id > LKST_MASKSET_ID_MAX) {
		/* Verify new maskset id. */
		ret = LKST_MASKSET_XCHG_ERRNO_ENID;
		LKST_STI(flags);
		return ret;
	}
	/* if new maskset has not been initialized,
	   return as error. */
	if (!LKST_POINTER_VERIFY
	    (lkst.maskset_table[new_id].masksetp, new_maskset)) {
		ret = LKST_MASKSET_XCHG_ERRNO_ENPT;
		LKST_STI(flags);
		return ret;
	}

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_MSET_XCHG,
		       LKST_ARG32(old_id, 0), LKST_ARG32(new_id, 0),
		       LKST_ARG32(old_maskset,0), LKST_ARG32(new_maskset, 0));

	LKST_MASKSET_SET_CURRENT(new_id);

	*id = old_id;

	if (xchg_hook)
		lkst_hook_maskset_xchg(new_id, old_id);

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_MSET_XCHG,
		       LKST_ARG32(old_id, 0), LKST_ARG32(new_id, 0),
		       LKST_ARG32(old_maskset,0), LKST_ARG32(new_maskset, 0));

	LKST_STI(flags);
	return ret;
}


/*
 * function lkst_evhandlerprim_daemon_wakeup_inline
 * A primitive part of an Event-handler. This awake a daemon process which
 * is waiting to read a specified buffer list.
 *
 *  Arguments
 *    cpu: cpu number coressponding to a buffer list..
 *
 *  returns
 *    NONE
 *
 */
static __inline__ void lkst_evhandlerprim_daemon_wakeup_inline(int cpu)
{
	if (cpu < smp_num_cpus && cpu >= 0) 
		lkst.owner_ready[cpu]=1;
}
void lkst_evhandlerprim_daemon_wakeup(int cpu)
{
	lkst_evhandlerprim_daemon_wakeup_inline(cpu);
}

/*********************************************************************
 * Event handle functions (embeded)
 *********************************************************************/

/* default event handler function */
void lkst_evhandler_default(void *phookrec, int event_type,
			    lkst_arg_t arg1,
			    lkst_arg_t arg2,
			    lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandlerprim_entry_log_l2_inline(event_type,
					       arg1, arg2, arg3, arg4);
}

/* default event handler control function */
int lkst_evhandler_default_ctrl(void *buf, size_t bufsize)
{
	return 0;
}


/* Event handler for buffer shifting */
void lkst_evhandler_buffer_shift(void *phookrec, int event_type,
				 lkst_arg_t arg1,
				 lkst_arg_t arg2,
				 lkst_arg_t arg3,
				 lkst_arg_t arg4)
{
	int	cpu = (int)arg1;
	lkst_buffer_id 	next = (lkst_buffer_id)arg2;

	if (next != LKST_BUFFER_ID_VOID)
		(void)lkst_evhandlerprim_buffer_shift_inline(cpu);
}


/* Event handler for maskset exchanging */
void lkst_evhandler_maskset_xchg(void *phookrec, int event_type,
				 lkst_arg_t arg1,
				 lkst_arg_t arg2,
				 lkst_arg_t arg3,
				 lkst_arg_t arg4)
{
	lkst_maskset_id id = (int) arg1;

	(void) lkst_evhandlerprim_maskset_xchg_inline(&id, 0);
}


/* Event handler for shifting buffer and awaking daemon process */
void lkst_evhandler_buffer_shift_dw(void *phookrec, int event_type,
				    lkst_arg_t arg1,
				    lkst_arg_t arg2,
				    lkst_arg_t arg3,
				    lkst_arg_t arg4)
{
	int cpu = (int) arg1;
	lkst_buffer_id 	next = (lkst_buffer_id) arg2;

	if (next != LKST_BUFFER_ID_VOID)
		(void) lkst_evhandlerprim_buffer_shift_inline(cpu);

	lkst_evhandlerprim_daemon_wakeup_inline(cpu);
}

#if DEBUG
/* 
 * event-handler-function lkst_evhandler_exception
 *
 * This event handler is used for exception event.
 * It can record multiple exception events, the first event is
 * written to currently selected buffer, and after second events are
 * written to stage0 buffer.
 *
 * This event handler is used as not to crush linux kernel because of
 * occurring multiple exception event.
 * 
 * "exhandler_addr" is an exception handler address.
 *
 */
void lkst_evhandler_exception(void *phookrec,
			      int event_type,
			      lkst_arg_t arg1,
			      lkst_arg_t arg2,
			      lkst_arg_t arg3,
			      lkst_arg_t arg4)
{
	static lkst_arg_t exhandler_addr = 0;
	int first = 0;
	struct lkst_event_buffer *buffer = NULL;
	struct lkst_event_record *event_entry_p;
	unsigned long flags;
	__volatile__ int recid;

	LKST_CLI(flags);

	/* exception handler address check */
	if (exhandler_addr != arg1) {
		first = 1;
		exhandler_addr = arg1;
		buffer = lkst_evhandlerprim_getw_buffer_inline();
	} else {
		/* get stage0 buffer pointer */
		buffer = lkst_evhandlerprim_select_buffer_id_inline(0);
	}

	recid = lkst_evhandlerprim_recid_inc_inline();
	event_entry_p = lkst_evhandlerprim_entry_next(buffer, event_type);

	if (event_entry_p)
		lkst_evhandlerprim_entry_log_l1(event_entry_p, recid,
						event_type, arg1, arg2, arg3, arg4);

	lkst_evhandlerprim_putw_buffer_inline(buffer);

	if (first) {
		first = 0;
		/* clear exception handler address */
		exhandler_addr = 0;
	}

	LKST_STI(flags);

	return;
}
#endif /* DEBUG */

/*********************************************************************
 * Lkst core functions
 *********************************************************************/
/*
 * This checks read mode of a specified buffer list.
 * If the current process is set to read the buffer list,
 * it returns 0. If not, it returns 1.
 * When the process is not set, that is not allowed to 
 * read, seek, create, delete buffers of the buffer list.
 */
static __inline__ int lkst_buffer_rmod_check(int cpu)
{
	if (lkst.buffer_owner[cpu] != current->pid) {
		return (lkst.buffer_owner[cpu] != (PID_MAX + 1));
	}
	return (0);
}

/*
 * This migrates caller process to specified CPU
 */
static __inline__ void lkst_cpu_migrate_set(int cpu)
{
	set_cpus_allowed(current,(unsigned long)1UL<<cpu);
}

/*
 * This resets migrated process
 */
static __inline__ void lkst_cpu_migrate_reset(void)
{
	current->cpus_allowed = -1;
}

static __inline__ int lkst_trc_status_core(struct lkst_status_param *param)
{
	int i;
	lkst_maskset *maskset;
	lkst_evhandler_func evhandler;
	struct lkst_event_buffer *buffer;

	/* !NOTE!: please check (param != NULL) before calling this.*/
	param->current_maskset_id = lkst.current_maskset_id;

	for(i = 0; i < LKST_CPU_MAX; i++)
		param->current_buffer_id[i] = lkst.current_buffer_id[i];

	param->maskset_num = 0;
	for(i = 0; i < LKST_MASKSET_TBL_MAX; i++)
		if (LKST_POINTER_VERIFY_STRICT(lkst.maskset_table[i].masksetp, maskset))
			param->maskset_num++;

	param->evhandler_num = 0;
	for(i = 0; i < LKST_EVHANDLER_TBL_MAX; i++)
		if (LKST_POINTER_VERIFY_STRICT(lkst.evhandler_table[i].evhandler, evhandler))
			param->evhandler_num++;

	param->buffer_num = 0;
	for(i = 0; i < LKST_BUFFER_TBL_MAX; i++)
		if (LKST_POINTER_VERIFY_STRICT(lkst.buffer_table[i].bufferp, buffer))
			param->buffer_num++;

	return 0;
}

/* lkst_trc_start* and lkst_trc_stop* functions are obsolete.*/
static __inline__ int lkst_trc_start_core(void)
{
	lkst_maskset *maskset;
	lkst_maskset_id current_id;

	spin_lock(&lkst_status_lock);

	if (lkst.maskset_id_saved > LKST_MASKSET_ID_MAX) {
		DPRINTK("Save maskset ID=%d is invalid!\n", lkst.maskset_id_saved);
	        lkst.maskset_id_saved = LKST_MASKSET_ID_RDEFAULT;
	}

	if (!LKST_POINTER_VERIFY(lkst.maskset_table[lkst.maskset_id_saved].masksetp, maskset)) {
		DPRINTK("Saved maskset ID=%d has already been deleted!\n", lkst.maskset_id_saved);
		lkst.maskset_id_saved = LKST_MASKSET_ID_RDEFAULT;
	}

	current_id = lkst.current_maskset_id;
	if (current_id != lkst.maskset_id_saved){
		LKST_MASKSET_SET_CURRENT(lkst.maskset_id_saved);
		lkst_hook_maskset_xchg(lkst.current_maskset_id, current_id);
	}

	spin_unlock(&lkst_status_lock);

	return (0);
}


static __inline__ int lkst_trc_stop_core(void)
{
	spin_lock(&lkst_status_lock);

	lkst.maskset_id_saved = lkst.current_maskset_id;

	lkst_hook_maskset_xchg(LKST_MASKSET_ID_RNOTHING, lkst.current_maskset_id);
	LKST_MASKSET_SET_CURRENT(LKST_MASKSET_ID_RNOTHING);

	spin_unlock(&lkst_status_lock);

	return (0);
}

static __inline__ int lkst_maskset_read_core(struct lkst_maskset_param *param)
{
	int i;
	lkst_maskset *maskset;
	struct lkst_maskset_entry *entry;
	lkst_evhandler_id id;
	size_t size;

	/* !NOTE!: please check (param != NULL) before calling this.*/
	size = param->maskset_size;

	if (param->id == LKST_MASKSET_ID_VOID)
		param->id = lkst.current_maskset_id;

	if (param->id > LKST_MASKSET_ID_MAX) {
		DPRINTK("Invalid maskset ID=%d\n", param->id);
		return -EINVAL;
	}

	atomic_inc(&(lkst.refcnt));

	if (!LKST_POINTER_VERIFY
	    (lkst.maskset_table[param->id].masksetp, maskset)) {
		DPRINTK("Maskset(ID=%d) does not exist!\n", param->id);
		atomic_dec(&(lkst.refcnt));
		return -EINVAL;
	}

	entry = param->maskset->entry;
	param->maskset->len = LKST_MASKSET_TABLE_LEN_MAX;

        size -= LKST_MASKSET_SIZE(0);

	for (i = 0; i < LKST_MASKSET_TABLE_LEN_MAX; i++) {
		id = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, i);
		entry->event_type = i;
		entry->id = id;
                size -= sizeof(struct lkst_maskset_entry);
                if (size < sizeof(struct lkst_maskset_entry)){
			param->maskset->len = i + 1;
			break;
                }
		entry++;
	}

	strncpy(param->maskset->name, lkst.maskset_table[param->id].name, (size_t) LKST_MASKSET_NAME_LEN);
	param->maskset->name[LKST_MASKSET_NAME_LEN - 1] = '\0';

	atomic_dec(&(lkst.refcnt));

	return 0;
}


static __inline__ int lkst_maskset_write_core(struct lkst_maskset_param *param)
{
	int i;
	int type;
	lkst_evhandler_id id;
	volatile lkst_maskset *maskset;
	lkst_evhandler_func evhandler;
	int lock = 0;
	int wakeup_daemon = 0;

	/* !NOTE!: please check (param != NULL) before calling this.*/
	if (param->id == LKST_MASKSET_ID_VOID) {
		/* search empty entry */
		for (i = 0; i < LKST_MASKSET_TBL_MAX; i++) {
			if (!LKST_POINTER_VERIFY
			    (lkst.maskset_table[i].masksetp, maskset))
				break;
		}
		if (i == LKST_MASKSET_TBL_MAX) {
			DPRINTK("Cannot register maskset any more!\n", "");
			return -EBUSY;
		}
		param->id = i;
	}

	if (param->id <= LKST_MASKSET_ID_RDEFAULT ||
	    param->id > LKST_MASKSET_ID_MAX) {
		DPRINTK("Invalid number of maskset ID=%d\n", param->id);
		return -EINVAL;
	}

	for (i = 0; i < param->maskset->len; i++ ) {
		type = param->maskset->entry[i].event_type;
		id = param->maskset->entry[i].id;

		if (type > LKST_ETYPE_MAX || type < 0) {
			DPRINTK("Invalid event_type=%d\n", type);
			return -EINVAL;
		}

		if (id > LKST_EVHANDLER_ID_MAX &&
		    id != LKST_EVHANDLER_ID_VOID) {
			DPRINTK("Invalid handler ID=%d\n", id);
			return -EINVAL;
		}
		if (!LKST_POINTER_VERIFY(lkst.evhandler_table[id].evhandler, evhandler) && id != LKST_EVHANDLER_ID_VOID) {
	    		DPRINTK("Handler(ID=%d) does not exist!\n", id);
			return -EINVAL;
		}

 		/* Check whether events concerned with spinlocks
		   are recorded. Event type of these events 
		   should be 0x8*.				  */
		if ((type >> 4) == (LKST_ETYPE_LK_SPINLOCK >> 4) &&
		    id != LKST_EVHANDLER_ID_VOID) 
			lock = 1;

		if (type == LKST_ETYPE_LKST_BUFF_OVFLOW &&
		    id == LKST_EVHANDLER_ID_BUFFER_SHIFT_DW)
			wakeup_daemon = 1;
	}

	if (lock && wakeup_daemon) {
		DPRINTK("Cannot use shift and wakeup event handler for buffer overflow event when recording lock events!\n", "");
		return -EINVAL;
	}		

	atomic_inc(&(lkst.refcnt));

	if (lkst.maskset_table[param->id].masksetp == NULL) {
		/* new maskset-table allocate */
		maskset = vmalloc(sizeof(lkst_evhandler_id) * (LKST_ETYPE_MAX + 1));
		if (maskset == NULL) {
			DPRINTK("Cannot allocate kernel memory!\n", "");
			atomic_dec(&(lkst.refcnt));
			return -ENOMEM;
		}

		if (!LKST_LIMIT_ADDRESS_CHECK(maskset, sizeof(lkst_evhandler_id) * (LKST_ETYPE_MAX + 1), lkst.limit_addr)) {
			DPRINTK("Cannot allocate maskset area! Aloocation limit address:0x%p, new maskset needs to 0x%p\n", lkst.limit_addr, (void *) ((char *) maskset + sizeof(lkst_evhandler_id) * (LKST_ETYPE_MAX + 1)));
			vfree((lkst_maskset *) maskset);
			atomic_dec(&(lkst.refcnt));
			return -ENOMEM;
		}

		/* init */
		for (i = 0; i <= LKST_ETYPE_MAX; i++)
			LKST_MASKSET_SET_EVHANDLER_ID(maskset, i, LKST_EVHANDLER_ID_VOID);

		LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table, param->id, 
					     param->maskset->name, maskset);

	} else {

		/* table will be overwritten */
		if (!LKST_POINTER_VERIFY
		    (lkst.maskset_table[param->id].masksetp, maskset)){
			DPRINTK("Maskset(ID=%d) is collapsed!\n", param->id);
			atomic_dec(&(lkst.refcnt));
			return -EBUSY;
		}

		strncpy(lkst.maskset_table[param->id].name,
			param->maskset->name,
			(size_t) LKST_MASKSET_NAME_LEN);
		lkst.maskset_table[param->id].name[LKST_MASKSET_NAME_LEN - 1] = '\0';
	}

	if(param->id == lkst.current_maskset_id)
	{
		for (i = 0; i < param->maskset->len; i++){
			lkst_event_type etype = param->maskset->entry[i].event_type;
			if((LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, etype) != param->maskset->entry[i].id)&&
			   (etype < LKST_MASKSET_TBL_MAX))
			{
				lkst_hook_exit_xchg(etype, param->maskset->entry[i].id, LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, etype));
			}
		}
	}

	/* main */
	for (i = 0; i < param->maskset->len; i++)
		LKST_MASKSET_SET_EVHANDLER_ID(maskset, param->maskset->entry[i].event_type, param->maskset->entry[i].id);

	atomic_dec(&(lkst.refcnt));

	return 0;
}


static __inline__ int lkst_maskset_xchg_core(lkst_maskset_id * id)
{
	int	ret;
	lkst_maskset_id old_id;

	/* !NOTE!: please check (id != NULL) before calling this.*/
	old_id= *id;

	ret = lkst_evhandlerprim_maskset_xchg_inline(id, 1);

	switch (ret) {
	case LKST_MASKSET_XCHG_ERRNO_SUCCESS:
		return 0;
	case LKST_MASKSET_XCHG_ERRNO_WCPT:
		DPRINTK("Current maskset(ID=%d) is collapsed!\n", old_id);
		return 0;
	case LKST_MASKSET_XCHG_ERRNO_WCID:
		DPRINTK("Invalid ID=%d of current maskset!\n", old_id);
		return 0;
	case LKST_MASKSET_XCHG_ERRNO_ECPT:
		DPRINTK("Current maskset(id=%d) is collapsed!\n", old_id);
		return -EBUSY;
	case LKST_MASKSET_XCHG_ERRNO_ENID:
		DPRINTK("Invalid maskset ID=%d\n", *id);
		return -EINVAL;
	case LKST_MASKSET_XCHG_ERRNO_ENPT:
		DPRINTK("Specified maskset(ID=%d) does not exist!\n", *id);
		return -EINVAL;
	case LKST_MASKSET_XCHG_ERRNO_ESTAT:
		DPRINTK("Cannot change maskset when LKST is not running!\n", "");
		return -EBUSY;
	default:
		DPRINTK("Unknown error occured, old/new maskset ID=%d/%d.\n", old_id, *id);
		return -EINVAL;
	}
}


static __inline__ int lkst_maskset_list_core(struct lkst_maskset_listparam *lp)
{
	int i, cnt, cnt_max;
	lkst_maskset *maskset;

	/* !NOTE!: please check (lp != NULL) before calling this.*/
	if(lp->listent_size < sizeof(struct lkst_maskset_listent) ||
	   lp->listent_size > LKST_MASKSET_LISTENT_SIZE(LKST_MASKSET_TBL_MAX)) {
		DPRINTK("Invalid listent_size=%d\n", lp->listent_size);
		return -EINVAL;
	}

	cnt = 0;
	cnt_max = lp->listent_size / sizeof(struct lkst_maskset_listent);

	for (i = 0; i < cnt_max; i++) {
		lp->listent[i].id = LKST_MASKSET_ID_VOID;
		lp->listent[i].name[0]='\0';
		lp->listent[i].len = 0;
	}
		
	for (i = 0; i < LKST_MASKSET_TBL_MAX; i++) {
		if (LKST_POINTER_VERIFY_STRICT(lkst.maskset_table[i].masksetp, maskset)) {

			lp->listent[cnt].id = i;
			strcpy(lp->listent[cnt].name, lkst.maskset_table[i].name);
			
			lp->listent[cnt].len = LKST_MASKSET_TABLE_LEN_MAX;
			cnt++;
			if(cnt > cnt_max)
				return 0;
		}
	}

	lp->current_id = lkst.current_maskset_id;

	return 0;
}


static __inline__ int lkst_maskset_delete_core(lkst_maskset_id id)
{
	lkst_maskset *maskset;

	if (id == LKST_MASKSET_ID_RALL ||
	    id == LKST_MASKSET_ID_RNOTHING ||
	    id == LKST_MASKSET_ID_RDEFAULT) {
		DPRINTK("Cannot delete special maskset!\n", "");
		return -EINVAL;
	}

	if (id > LKST_MASKSET_ID_MAX) {
		DPRINTK("Invalid maskset ID=%d\n", id);
		return -EINVAL;
	}

	if (id == lkst.current_maskset_id) {
		DPRINTK("Specified maskset(ID=%d) is currently used!\n", id);
		return -EBUSY;
	}
	
	/* for obsolete routines:-P */
	if (id == lkst.maskset_id_saved) {
		DPRINTK("Specified maskset(ID=%d) is currently saved!\n Change to DEFAULT maskset.", id);
		lkst.maskset_id_saved = LKST_MASKSET_ID_RDEFAULT;
	}

	if (!LKST_POINTER_VERIFY
	    (lkst.maskset_table[id].masksetp, maskset)) {
		DPRINTK("Maskset(ID=%d) does not exist!\n", id);
		return -EINVAL;
	}

	LKST_POINTER_INVALIDATE(lkst.maskset_table[id].masksetp);

			/* Wait until all the event handlers exit. */
	while (atomic_read(&(lkst.refcnt)));

			/* Reconfirm this maskset is not used currently. */
	if (id == lkst.current_maskset_id) {
		DPRINTK("Specified maskset(ID=%d) is currently used!\n", id);

			/* Restore buffer pointer */
		lkst.maskset_table[id].masksetp = maskset;
		LKST_POINTER_CHECK_INIT(lkst.maskset_table[id].masksetp);

		return -EBUSY;
	}

	vfree(maskset);

	return 0;
}

static __inline__ int 
lkst_evhandler_list_core(struct lkst_evhandler_listparam *lp)
{
	int i, cnt, cnt_max;
	lkst_evhandler_func evhandler;

	/* !NOTE!: please check (lp != NULL) before calling this.*/
	if(lp->listent_size < sizeof(struct lkst_evhandler_listent) ||
	   lp->listent_size > LKST_EVHANDLER_LISTENT_SIZE(LKST_EVHANDLER_TBL_MAX)) {
		DPRINTK("Invalid listent_size=%d\n", lp->listent_size);
		return -EINVAL;
	}

	cnt = 0;
	cnt_max = lp->listent_size / sizeof(struct lkst_evhandler_listent);

	for (i = 0; i < cnt_max; i++) {
		lp->listent[i].id = LKST_EVHANDLER_ID_VOID;
		strcpy(lp->listent[i].name, "");
	}
		
	for (i = 0; i < LKST_EVHANDLER_TBL_MAX; i++) {
		if (LKST_POINTER_VERIFY(lkst.evhandler_table[i].evhandler, evhandler)) {

			lp->listent[cnt].id = i;
			strcpy(lp->listent[cnt].name, lkst.evhandler_table[i].name);
			
			cnt++;
			if(cnt > cnt_max)
				return 0;
		}
	}

	return 0;
}


static __inline__ int 
lkst_evhandler_ctrl_core(struct lkst_evhandler_ctrl_param *evhandler_ctrl_param)
{
	lkst_evhandler_id id;
	lkst_evhandler_ctrl_func evhandler_ctrl;

	id = evhandler_ctrl_param->id;
	if (id > LKST_EVHANDLER_ID_MAX) {
		DPRINTK("Invalid handler ID = %d\n", id);
		return -EINVAL;
	}

	if (!LKST_POINTER_VERIFY(lkst.evhandler_table[id].evhandler_ctrl,
				 evhandler_ctrl)) {
		DPRINTK("Ctrl-function(ID=%d) does not exist!\n", id);
		return -EINVAL;
	}

	evhandler_ctrl_param->ret = (*evhandler_ctrl) (evhandler_ctrl_param->buf,
						       evhandler_ctrl_param->bufsize);


	return 0;
}

/*
 * lkst_evhandler_register_core
 * return value : 
 * 	success 0
 *	fail	errno
 * Called from lkst_evhandler_register, lkst_evhandler_unregister_core, 
 * lkst_init_stage0
 * This function do !NOT! check which name is used. please check it before call.
 */
static __inline__ int 
lkst_evhandler_register_core(lkst_evhandler_id id, const char *name,
			     lkst_evhandler_func evhandler,
			     lkst_evhandler_ctrl_func evhandler_ctrl)
{
	if ( name != NULL ) {
		strncpy(lkst.evhandler_table[id].name, name,
			(size_t) LKST_EVHANDLER_NAME_LEN);
		lkst.evhandler_table[id].name[LKST_EVHANDLER_NAME_LEN - 1] = '\0';
	} else
		lkst.evhandler_table[id].name[0] = '\0';

	lkst.evhandler_table[id].evhandler = evhandler;
	if (evhandler != NULL)
		LKST_POINTER_CHECK_INIT(lkst.evhandler_table[id].evhandler);
	else
		lkst.evhandler_table[id].evhandler_check = 0;
	
	lkst.evhandler_table[id].evhandler_ctrl = evhandler_ctrl;
	if (evhandler_ctrl != NULL)
		LKST_POINTER_CHECK_INIT(lkst.evhandler_table[id].
					evhandler_ctrl);
	else
		lkst.evhandler_table[id].evhandler_ctrl_check = 0;
	
	return lkst_hook_exit_update(id);
}

/*
 * lkst_evhandler_unregister_core
 * return value :
 *	success	0
 *	fail	errno
 * Called from lkst_evhandler_unregister, lkst_trc_init_stage0, lkst_cleanup_module
 */
static __inline__ int lkst_evhandler_unregister_core(lkst_evhandler_id id)
{
	return lkst_evhandler_register_core(id, NULL, NULL, NULL);
}

/*
 * lkst_evhandler_get_id_core
 * return value : evhandler_id
 * Called from lkst_evhandler_register,lkst_evhandler_get_id, lkst_evhandler_get_id_ioctl(not implemented)
 */
static __inline__ lkst_evhandler_id
lkst_evhandler_get_id_core(const char *name)
{
	lkst_evhandler_id id;
	/* !NOTE!: please check (name != NULL) && (strlen(name)!=0) before calling.*/
	for ( id = 0; id <= LKST_EVHANDLER_ID_MAX; id++) {
		if (lkst.evhandler_table[id].evhandler != NULL) {
			if (strncmp(name, 
				    lkst.evhandler_table[id].name,
				    LKST_EVHANDLER_NAME_LEN) == 0){
				return id;
			}
		}
	}
	return LKST_EVHANDLER_ID_VOID;
}


/*
 * lkst_write_to_log
 * return value : recid
 * write down event-buffer entry to log-buffer.
 */
static void lkst_write_to_log(struct lkst_event_record *entry,
			      struct lkst_log_record *log)
{
	struct task_struct *task = NULL;
	static uid_t uid;
	static gid_t gid;
	static pid_t pid = PID_MAX + 1;
	static pid_t pgrp;

	/* Is the current (logged) pid same as the last?*/
	if (pid != entry->log_pid) {
		/*Different! hum, then, we have to check that uid, gid and pgrp.*/
		pid = entry->log_pid;
		read_lock(&tasklist_lock);
		task = find_task_by_pid(pid);
		if (task == NULL) {
		  /* write dummy value if task_struct was already deleted */
			uid = -1;	// TODO: write after
			gid = -1;	// TODO: write after
			pgrp = -1;	// TODO: write after
		} else {
			uid = task->uid;
			gid = task->gid;
			pgrp = task->pgrp;
		}
		read_unlock(&tasklist_lock);
	}

	log->posix.log_magic = 0x76758384;	// TODO: value
	log->posix.log_size = sizeof(lkst_arg_t) * 4;
	log->posix.log_format = PXLOG_BINARY;
	log->posix.log_facility = LOG_KERN;
	log->posix.log_severity = LOG_DEBUG;
	log->posix.log_recid = entry->log_recid;
	log->posix.log_event_type = entry->log_event_type;
	log->posix.log_pid = pid;
	log->posix.log_uid = uid;
	log->posix.log_gid = gid;
	log->posix.log_pgrp = pgrp;
	log->posix.log_flags = 0;
	log->posix.log_time.tv_sec = (time_t) ((entry->log_time) >> 32);
	log->posix.log_time.tv_nsec = (unsigned long) entry->log_time;
	log->posix.log_thread = -1;		// TODO: write after
	log->log_arg1 = entry->log_arg1;
	log->log_arg2 = entry->log_arg2;
	log->log_arg3 = entry->log_arg3;
	log->log_arg4 = entry->log_arg4;
}

static __inline__ int lkst_buffer_read_core(struct lkst_log_buffer *lbuffer,
			unsigned char flag)
{
	size_t readp, readp_s, writep;
	size_t count, rsize, result_readp;
	size_t log_writep;
	size_t sizem;
	int cpu;
	int wrapped = 0;
	int current_f;
	int ring;
	struct lkst_event_buffer *buffer=NULL;
	struct lkst_event_record *entry, *prev_entry;
	struct lkst_log_record record, *record_p;
	lkst_buffer_id id;
	int prev_recid;
	int rcnt, rwdiff;

	/* !NOTE!: please check (lbuffer != NULL) before calling.*/
	count = lbuffer->read_size;
	lbuffer->result_read_size = 0;

	if (lbuffer->id > LKST_BUFFER_ID_MAX) {
		DPRINTK("Invalid buffer ID = %d\n", lbuffer->id);
		return -EINVAL;
	}

	id = lbuffer->id;

	log_writep = 0;
	if (flag)
		/* Called from lkst_buffer_read_ioctl() or lkst_read() */
		record_p = &record;
	else
		/* Called from lkst_buffer_read() (for module use) */
		record_p = lbuffer->buffer;

	// atomic_inc(&(lkst.refcnt)); CPU migrate

#if DEBUG
retry:
#endif /* DEBUG */

	if (!LKST_POINTER_VERIFY(lkst.buffer_table[id].bufferp, buffer)) {
		DPRINTK("Buffer(ID=%d) does not exist!\n", id);
		// atomic_dec(&(lkst.refcnt)); CPU migrate
		/* return with result_read_size = 0. */
		return -EINVAL;
	}

	cpu = buffer->cpu;

	/* CPU binding */
	if (!(flag & 0x02)) {
		lkst_cpu_migrate_set(cpu);
		if (!LKST_POINTER_VERIFY(lkst.buffer_table[id].bufferp,
					 buffer)) {
			DPRINTK("Buffer(ID=%d) does not exist!\n", id);
			lkst_cpu_migrate_reset();
			return -EINVAL;
		}
	} else {
		rcnt = buffer->rcnt.counter; // atomic_read(&(buffer->rcnt)); CPU migrate
	}

	ring = lkst.buffer_ring[cpu];

	if ((flag & 0x01) && ring && id != 0) {
		DPRINTK("Cannot read ring structure buffer!\n", "");
		// atomic_dec(&(lkst.refcnt)); CPU migrate
		/* return with result_read_size = 0. */
		if (!(flag & 0x02)) lkst_cpu_migrate_reset();
		return -EBUSY;
	}		

	if (lkst_buffer_rmod_check(cpu)) {
		DPRINTK("Process(pid=%d) is not set to read this buffer list(cpu=%d).\n", current->pid, cpu);
		// atomic_dec(&(lkst.refcnt)); CPU migrate
		/* return with result_read_size = 0. */
		if (!(flag & 0x02)) lkst_cpu_migrate_reset();
		return -EBUSY;
	}		

	sizem = buffer->size - 1;

	/* get buffer time stamp */
	/* TODO : buffer's timestamp should be modified when lkst_header is modified.*/
	if (!lkst.cmos_time_flag && buffer->cmos_time_flag) {
		buffer->xtime.tv_sec = lkst.xtime.tv_sec;
		buffer->xtime.tv_usec = lkst.xtime.tv_usec;
		buffer->cmos_time_flag = 0;
	}
	lbuffer->tsc = lkst.tsc;
	lbuffer->xtime.tv_sec = buffer->xtime.tv_sec;
	lbuffer->xtime.tv_usec = buffer->xtime.tv_usec;
	lbuffer->cpu_freq = lkst.cpu_freq;
	lbuffer->buf_ver = lkst.buf_ver;
	strncpy(lbuffer->arch, lkst.arch, LKST_ARCH_NAME_LEN);
	lbuffer->endian_big = lkst.endian_big;

	/* exit if count is smaller than the size of one log entry */
	if (count < sizeof(struct lkst_log_record)) {
		/* return with result_read_size = 0. */
		// atomic_dec(&(lkst.refcnt)); CPU migrate
		if (!(flag & 0x02)) lkst_cpu_migrate_reset();
		return 0;
	}

	readp = buffer->readp;
	current_f = (lkst.current_buffer_id[cpu] == id);

	/* load previous entry of read pointer */
	(char *) prev_entry = (char *) (buffer->body) +
		(buffer->last_readp & sizem);

	/* get buffer write pointer */
	writep = (size_t) (buffer->writep.counter & sizem);

	/* buffer wrap check */
	if (buffer->last_recid != prev_entry->log_recid) {
		wrapped = 1;
		readp = writep;
		(char *) prev_entry = (char *) (buffer->body) + 
			((readp - sizeof(struct lkst_event_record)) & sizem);
	} else if (readp == writep) {
		/* buffer is empty */
		/* return with result_read_size = 0. */
		// atomic_dec(&(lkst.refcnt)); CPU migrate
		if (!(flag & 0x02)) lkst_cpu_migrate_reset();
		return 0;
	}

	/* HEADSPACE read check */
	if (ring) {
		rsize = count / sizeof(struct lkst_event_record) *
			sizeof(struct lkst_log_record);

		result_readp = (rsize + readp) & sizem;

		if (rsize < buffer->size &&
		    result_readp < LKST_BUFFER_HEADSPACE) {
			if (rsize > result_readp) {
				rsize -= result_readp;
				count -= LKST_BUFFER_SIZE_LOCAL2EXT(result_readp);
			} else {
				DPRINTK("Cannot read buffer headspace, please read larger size.\n", "");
				// atomic_dec(&(lkst.refcnt)); CPU migrate
				if (!(flag & 0x02)) lkst_cpu_migrate_reset();
				return -EINVAL;
			}
		}
	}

	/* set read stopper */
	if (!current_f)
		readp_s = writep;
	else 
		readp_s = readp;

	(char *) entry = (char *) (buffer->body) + (readp & sizem);

#if DEBUG
		/* Assume no next buffer to read. */
	id = LKST_BUFFER_ID_VOID;
#endif /* DEBUG */

	if (wrapped)
		prev_recid = entry->log_recid;
	else
		prev_recid = prev_entry->log_recid;

	while (log_writep + sizeof(struct lkst_log_record) <= count) {

		lkst_write_to_log(entry, record_p);
		record_p->posix.log_processor = (posix_log_procid_t) cpu;

		/* entry overwritten check */
		if (record_p->posix.log_recid <= prev_recid) {
			record_p->log_arg4 = LKST_ARG32(record_p->posix.log_event_type, 0);
			record_p->posix.log_event_type = LKST_ETYPE_LKST_OVWRTN_REC;

				/* flag & 0x2 :Called from lkst_read() */
			if (flag & 0x02 && buffer->next != LKST_BUFFER_ID_VOID) { 
				/* buffer read/write count check */
				rwdiff = atomic_read(&(buffer->wcnt)) - rcnt;
				if (rwdiff >= 2) {
					/* Get next buffer ID */
					lbuffer->id = buffer->next;
					break;
				}
			}

			if (current_f && record_p->posix.log_recid != 0)
				wrapped = 1;
		}
		    
		prev_entry = entry;
		/* update prev_recid */
		if (prev_recid < prev_entry->log_recid || !ring)
			prev_recid = prev_entry->log_recid;

		entry++;

		if (flag) {
			if (copy_to_user((char *) (lbuffer->buffer) + log_writep, &record, sizeof(struct lkst_log_record))) {
				DPRINTK("Writing fail at user memory address=%p\n", (lbuffer->buffer) + log_writep);
				break;
			} 
		} else
			record_p++;

		buffer->last_recid = record_p->posix.log_recid;
		buffer->last_readp = readp & sizem;

		log_writep += sizeof(struct lkst_log_record);

		readp += sizeof(struct lkst_event_record);

		if (readp >= buffer->size) {
			entry = buffer->body;
			readp &= sizem;
		}

                if (current_f && !wrapped) {
                        writep = atomic_read(&(buffer->writep)) & sizem;
                        if (readp == writep)
                                break;
                }

		if (readp == readp_s) {
				/* flag & 0x2 :Called from lkst_read() */
			if ((flag & 0x2) && buffer->next != LKST_BUFFER_ID_VOID) {
#if DEBUG
				/* Remember next buffer to chain reading. */
				id = buffer->next;
#endif /* DEBUG */
				/* Get next buffer ID */
				lbuffer->id = buffer->next;
			}
			break;
		}
	}

	buffer->readp = readp;

	lbuffer->result_read_size = log_writep;
#if DEBUG
	if (id != LKST_BUFFER_ID_VOID) {
		// atomic_dec(&(lkst.refcnt)); CPU migrate
				/* Wait next buffer filled if needed. */
		// atomic_inc(&(lkst.refcnt)); CPU migrate
		goto retry;
	}
#endif /* DEBUG */

	// atomic_dec(&(lkst.refcnt)); CPU migrate

	/* CPU binding */
	if (!(flag & 0x02)) lkst_cpu_migrate_reset();

	return 0;
}

#if DEBUG
static __inline__ int lkst_buffer_seek_core(struct lkst_buffer_seekparam *bs)
{
	struct lkst_event_buffer *buffer;
	struct lkst_event_record *prev_entry;
	size_t buff_size, writep, readp;
	int offset, diff;

	/* !NOTE!: please check (bs != NULL) before calling this.*/
	if (bs->id > LKST_BUFFER_ID_MAX) {
		DPRINTK("Invalid buffer ID=%d\n", bs->id);
		return -EINVAL;
	}

	offset = bs->offset;

	if (offset % ((int) sizeof(struct lkst_event_record))) {
		offset -= (offset % ((int) sizeof(struct lkst_event_record)));
		DPRINTK("Offset must be multiple of buffer entry size(%d), modified to %d\n", sizeof(struct lkst_event_record), offset);
	}

	spin_lock(&lkst_buffer_lock);

	if (!LKST_POINTER_VERIFY(lkst.buffer_table[bs->id].bufferp, buffer)) {
		DPRINTK("Buffer(ID=%d) does not exist!\n", bs->id);
		spin_unlock(&lkst_buffer_lock);
		return -EINVAL;
	}

	if (lkst_buffer_rmod_check(buffer->cpu)) {
		DPRINTK("Process(pid=%d) is not allowed to control this buffer list(cpu=%d).\n", current->pid, buffer->cpu);
		spin_unlock(&lkst_buffer_lock);
		return -EBUSY;
	}		

	buff_size = buffer->size;

	if (offset > (int) buff_size || offset < -((int) buff_size)) {
		DPRINTK("Invalid offset=%d!\n", offset);
		spin_unlock(&lkst_buffer_lock);
		return -EINVAL;
	}

	readp = buffer->readp;

	(char *) prev_entry = (char *) (buffer->body) + buffer->last_readp;

	/* wrap check */
	if (buffer->last_recid == prev_entry->log_recid) {
		/* not wrapped */
		writep = (atomic_read(&(buffer->writep)) &
			  (buff_size - 1));

		diff = readp <= writep ?
			writep - readp : writep + buff_size - readp;

		if (offset >= diff)
			readp = writep;
		else if (offset <= (diff - (int) buff_size))
			readp = (writep + sizeof(struct lkst_event_record))
				& (buff_size - 1);
		else
			readp = (readp + offset) & (buff_size - 1);

		prev_entry = 
			LKST_BUFFER_PREV_ENTRY(buffer->body, readp, buff_size);

		buffer->last_recid = prev_entry->log_recid;
	} else {
		/* wrapped */
		readp = (readp + offset) & (buff_size - 1);
	}

	buffer->readp = readp;
	buffer->last_readp = 
		(readp - sizeof(struct lkst_event_record)) & (buff_size - 1);

	bs->result_readp = readp;

	spin_unlock(&lkst_buffer_lock);

	return 0;
}
#endif /* DEBUG */

static __inline__ int 
lkst_buffer_create_core_cpu(int cpu,
			    struct lkst_buffer_param *param)
{
	lkst_buffer_id prev_id;
	struct lkst_event_buffer *buffer, *prev_buffer = NULL, *head_buffer;
	size_t size;
	int i;

	/* !NOTE!: please check (param != NULL) before calling this.*/
	if (param->id == LKST_BUFFER_ID_VOID) {
		/* search empty entry */
		for (i = 0; i < LKST_BUFFER_TBL_MAX; i++) {
			if (!LKST_POINTER_VERIFY
			    (lkst.buffer_table[i].bufferp, buffer))
				break;
		}
		if (i == LKST_BUFFER_TBL_MAX) {
			DPRINTK("Cannot create buffer any more!\n", "");
			return -EBUSY;
		}
		param->id = i;
	}


	if (param->id > LKST_BUFFER_ID_MAX) {
		DPRINTK("Invalid buffer ID = %d\n", param->id);
		return -EINVAL;
	}

	if (LKST_POINTER_VERIFY(
		lkst.buffer_table[param->id].bufferp, buffer)) {
		DPRINTK("Buffer(ID=%d) already exist!\n", param->id);
		return -EINVAL;
	}

	if (param->size < LKST_BUFFER_SIZE_MIN) {
		DPRINTK("buffer size(=%d) is too small, changed to %d\n", param->size, LKST_BUFFER_SIZE_MIN);
		param->size = LKST_BUFFER_SIZE_MIN;
	} else if (param->size > LKST_BUFFER_SIZE_MAX) {
		DPRINTK("buffer size(=%d) is too large, changed to %d\n", param->size, LKST_BUFFER_SIZE_MAX);
		param->size = LKST_BUFFER_SIZE_MAX;
	}

	for (size = LKST_BUFFER_SIZE_MAX; size > param->size; size = size >> 1);
	if (param->size != size)
		DPRINTK("Buffer size must be power of 2, changed to %d\n", size);

	param->result_size = size;

	/* create new buffer area */
	buffer = vmalloc(LKST_BUFFER_AREA_SIZE(size));
	if (buffer == NULL) {
		DPRINTK("Cannot allocate kernel memory(size=%d)\n", LKST_BUFFER_AREA_SIZE(size));
		return -ENOMEM;
	}
	if (!LKST_LIMIT_ADDRESS_CHECK(buffer, size, lkst.limit_addr)) {
		DPRINTK("Cannot allocate buffer area! Aloocation limit address:0x%p, new buffer needs to 0x%p\n", lkst.limit_addr, (void *) ((char *) buffer + size));
		vfree(buffer);
		return -ENOMEM;
	}
	
	spin_lock(&lkst_buffer_lock);

			/* Wait until all the event handlers exit. */
	while (atomic_read(&(lkst.refcnt)));

	/* acquire buffer_tail */
	prev_id = lkst.buffer_tail[cpu];

	if (prev_id != LKST_BUFFER_ID_VOID) {
		if (!LKST_POINTER_VERIFY
		    (lkst.buffer_table[prev_id].bufferp, prev_buffer)) {
			DPRINTK("Buffer ID=%d of the specified cpu is collapsed! (cannot connect new buffer to the buffer list)\n", prev_id);
			vfree(buffer);
			spin_unlock(&lkst_buffer_lock);
			return -EBUSY;
		}
		buffer->next = prev_buffer->next;
	} else 
		buffer->next = LKST_BUFFER_ID_VOID;

	if (lkst.buffer_ring[cpu]) {
		/* This buffer is in ring structure */

		if (!LKST_POINTER_VERIFY
		    (lkst.buffer_table[buffer->next].bufferp, head_buffer)) {
			DPRINTK("Buffer ID=%d is collapsed!\n", buffer->next);
			vfree(buffer);
			spin_unlock(&lkst_buffer_lock);
			return -EBUSY;
		}
		head_buffer->prev = param->id;
	}

	buffer->size = size;

	buffer->readp = LKST_BUFFER_HEADSPACE;
	atomic_set(&(buffer->writep), LKST_BUFFER_HEADSPACE);

	buffer->id = param->id;
	buffer->prev = prev_id;
	buffer->cpu = cpu;
	buffer->last_recid = 0;
	buffer->last_readp = 
		(LKST_BUFFER_HEADSPACE - sizeof(struct lkst_event_record)) &
		(size - 1);

	buffer->cmos_time_flag = lkst.cmos_time_flag;
	buffer->xtime.tv_sec = lkst.xtime.tv_sec;
	buffer->xtime.tv_usec = lkst.xtime.tv_usec;

	strncpy(buffer->eyecatcher, "LKBF", 4);
	buffer->buf_ver = LKST_BUF_VER;

	atomic_set(&(buffer->rcnt), 0);
	atomic_set(&(buffer->wcnt), 0);

	/* clear recid, event_type */
	for (i = 0; i < (size / sizeof(struct lkst_event_record)); i++) {
		(buffer->body + i)->log_recid = 0;
		(buffer->body + i)->log_event_type = LKST_ETYPE_VOID;
	}

	/* modify buffer_tail */
	lkst.buffer_tail[cpu] = param->id;

	/* register to buffer_table */
	lkst.buffer_table[param->id].bufferp = buffer;
	LKST_POINTER_CHECK_INIT(lkst.buffer_table[param->id].bufferp);

	/* connect to tail of buffer list of specified cpu */
	if (prev_id != LKST_BUFFER_ID_VOID)
		prev_buffer->next = param->id;

	spin_unlock(&lkst_buffer_lock);

	return 0;
}


static __inline__ int lkst_buffer_create_core(struct lkst_buffer_param *param)
{
	int cpu, cpu_count = 1;
	int i;
	int retval;

	/* !NOTE!: please check (param != NULL) before calling this.*/
	cpu = param->cpu;

	if (cpu == -1) { /* create buffers for all CPUs */
		cpu = 0;
		cpu_count = smp_num_cpus;
	} else if (cpu >= smp_num_cpus || cpu < 0) {
		DPRINTK("Invalid CPU number=%d\n", cpu);
		return -EINVAL;
	}

	for (i = 0; i < cpu_count; i++) {
		if (param->cpu == -1)
			param->id = LKST_BUFFER_ID_VOID;
		if (lkst_buffer_rmod_check(cpu)) {
			DPRINTK("Process(pid=%d) is not allowed to control this buffer list(cpu=%d).\n", current->pid, cpu);
			return -EBUSY;
		}
		retval = lkst_buffer_create_core_cpu(cpu, param);
		if (retval)
			return retval;
		cpu++;
	}

	return 0;
}


static __inline__ int lkst_buffer_shift_core(int cpu)
{
	int	ret;

	if (lkst.buffer_ring[cpu]) {
		DPRINTK("Cannot shift ring structure buffer!\n", "");
		return -EBUSY;
	}

	if (lkst_buffer_rmod_check(cpu)) {
		DPRINTK("Process(pid=%d) is not allowed to control this buffer list(cpu=%d).\n", current->pid, cpu);
		return -EBUSY;
	}		

	ret = lkst_evhandlerprim_buffer_shift_inline(cpu);

	switch (ret) {
	case LKST_BUFFER_SHIFT_ERRNO_SUCCESS:
		return 0;
	case LKST_BUFFER_SHIFT_ERRNO_ECPU:
	  	DPRINTK("Invalid CPU number=%d\n", cpu); 
		return -EINVAL;
	case LKST_BUFFER_SHIFT_ERRNO_ECBUF:
		DPRINTK("Current buffer of CPU=%d does not exist!\n", cpu); 
		return -EBUSY;
	case LKST_BUFFER_SHIFT_ERRNO_ENBUF:
		DPRINTK("Next buffer of CPU=%d does not exist!\n", cpu);
		return -EBUSY;
	case LKST_BUFFER_SHIFT_ERRNO_EBUFC:
		DPRINTK("Next buffer of CPU=%d seems to be in deleting!\n", cpu);
		return -EBUSY;
	default:
		DPRINTK("Unknown error occured, CPU=%d.\n", cpu);
		return -EINVAL;
	}
}


static __inline__ int lkst_buffer_delete_core(lkst_buffer_id id)
{
	struct lkst_event_buffer *buffer, *prev_buffer, *next_buffer;

	if (id == 0) {
		DPRINTK("Cannot delete buffer of ID=%d\n", id);
		return -EINVAL;
	}

	// spin_lock(&lkst_buffer_lock); CPU migrate

	if (!LKST_POINTER_VERIFY_STRICT(lkst.buffer_table[id].bufferp, buffer)) {
		DPRINTK("Buffer(ID=%d) does not exist!\n", id);
		// spin_unlock(&lkst_buffer_lock); CPU migrate
		return -EINVAL;
	}

	/* CPU binding */
	lkst_cpu_migrate_set(buffer->cpu);
	if (!LKST_POINTER_VERIFY_STRICT(lkst.buffer_table[id].bufferp,
					buffer)) {
		DPRINTK("Buffer(ID=%d) does not exist!\n", id);
		lkst_cpu_migrate_reset();
		return -EINVAL;
	}

	if (lkst_buffer_rmod_check(buffer->cpu)) {
		DPRINTK("Process(pid=%d) is not allowed to control this buffer list(cpu=%d).\n", current->pid, buffer->cpu);
		// spin_unlock(&lkst_buffer_lock); CPU migrate
		lkst_cpu_migrate_reset();
		return -EBUSY;
	}		

			/* Check whether this buffer is used currently. */
	if (lkst.current_buffer_id[buffer->cpu] == id) {
		DPRINTK("Specified buffer(ID=%d) is currently used!\n", id);
		// spin_unlock(&lkst_buffer_lock); CPU migrate
		lkst_cpu_migrate_reset();
		return -EBUSY;
	}

			/* Invalidate buffer pointer */
	LKST_POINTER_INVALIDATE(lkst.buffer_table[id].bufferp);

			/* Wait until all the event handlers exit. */
	// while (atomic_read(&(lkst.refcnt))); CPU migrate

			/* Reconfirm this buffer is not used currently. */
	if (lkst.current_buffer_id[buffer->cpu] == id) {
		DPRINTK("Specified buffer(ID=%d) is currently used!\n", id);

			/* Restore buffer pointer */
		lkst.buffer_table[id].bufferp = buffer;
		LKST_POINTER_CHECK_INIT(lkst.buffer_table[id].bufferp);

		// spin_unlock(&lkst_buffer_lock); CPU migrate
		lkst_cpu_migrate_reset();
		return -EBUSY;
	}

	if (buffer->prev != LKST_BUFFER_ID_VOID) {
		if (!LKST_POINTER_VERIFY_STRICT
		    (lkst.buffer_table[buffer->prev].bufferp,
		     prev_buffer)) {
			DPRINTK("Previous buffer(ID=%d) is collapsed!\n", buffer->prev);
		} else {
			/* change next ID of previous buffer */
			prev_buffer->next = buffer->next;
		}
	}
	else if (lkst.buffer_head[buffer->cpu] != id) { // for debug
		DPRINTK("Buffer list. (buffer_head[%d] of event_buffer) is collapsed!\n", buffer->cpu);
	}

	if (buffer->next != LKST_BUFFER_ID_VOID) {
		if (!LKST_POINTER_VERIFY_STRICT
		    (lkst.buffer_table[buffer->next].bufferp,
		     next_buffer)) {
			DPRINTK("Next buffer(ID=%d) is collapsed!\n", buffer->next);
		} else {
			/* chane prev ID of next buffer */
			next_buffer->prev = buffer->prev;
		}
	}
	else if (lkst.buffer_tail[buffer->cpu] != id) { // for debug
		DPRINTK("Buffer list. (buffer_tail[%d] of event_buffer) is collapsed!\n", buffer->cpu);
	}

	if (lkst.buffer_tail[buffer->cpu] == id) {
		/* modify buffer_tail */
		lkst.buffer_tail[buffer->cpu] = buffer->prev;
	}

	if (lkst.buffer_head[buffer->cpu] == id) {
		/* modify buffer_head */
		lkst.buffer_head[buffer->cpu] = buffer->next;
	}

	/* free buffer area */
	vfree(buffer);

	// spin_unlock(&lkst_buffer_lock); CPU migrate

	/* CPU binding */
	lkst_cpu_migrate_reset();

	return (0);
}


static __inline__ int lkst_buffer_list_core(struct lkst_buffer_listparam *lp)
{
	int i, cnt, cnt_max;
	size_t buff_size;
	struct lkst_event_buffer *buffer;
	struct lkst_event_record *prev_entry;

	/* !NOTE!: please check (lp != NULL) before calling this.*/
	if(lp->listent_size < sizeof(struct lkst_buffer_listent) ||
	   lp->listent_size > LKST_BUFFER_LISTENT_SIZE(LKST_BUFFER_TBL_MAX)) {
		DPRINTK("Invalid listent_size=%d\n", lp->listent_size);
		return -EINVAL;
	}

	cnt = 0;
	cnt_max = lp->listent_size / sizeof(struct lkst_buffer_listent);

	for (i = 0; i < cnt_max; i++) {
		lp->listent[i].id = LKST_BUFFER_ID_VOID;
		lp->listent[i].size = 0;
		lp->listent[i].readp = 0;
		lp->listent[i].writep = 0;
		lp->listent[i].wrap_flag = LKST_FLAG_NO;
		lp->listent[i].prev = 0;
		lp->listent[i].next = 0;
		lp->listent[i].cpu = -1;
	}

	for (i = 0; i < LKST_BUFFER_TBL_MAX; i++) {
		spin_lock(&lkst_buffer_lock);
		if (!LKST_POINTER_VERIFY_STRICT(lkst.buffer_table[i].bufferp, buffer)) {
			spin_unlock(&lkst_buffer_lock);
			continue;
		}

		buff_size = buffer->size;

		lp->listent[cnt].id = buffer->id;
		lp->listent[cnt].size = buff_size;
		lp->listent[cnt].readp = buffer->readp;
		lp->listent[cnt].prev = buffer->prev;
		lp->listent[cnt].next = buffer->next;
		lp->listent[cnt].cpu = buffer->cpu;
		lp->listent[cnt].current_flag = 
			(lkst.current_buffer_id[buffer->cpu] == buffer->id)? LKST_FLAG_YES : LKST_FLAG_NO;
		lp->listent[cnt].writep = 
			atomic_read(&(buffer->writep)) & (buff_size - 1);

		(char *) prev_entry =
			(char *) (buffer->body) + buffer->last_readp;

		if (buffer->last_recid != prev_entry->log_recid) 
			lp->listent[cnt].wrap_flag = LKST_FLAG_YES;

		spin_unlock(&lkst_buffer_lock);

		cnt++;
		if (cnt > cnt_max) return 0;
	}

	return 0;
}


static __inline__ int lkst_buffer_ring_core(struct lkst_buffer_ringparam *param)
{
	lkst_buffer_id head_id, tail_id;
	struct lkst_event_buffer *head_buffer, *tail_buffer;

	/* !NOTE!: please check (param != NULL) before calling this.*/
	if (param->cpu >= smp_num_cpus || param->cpu < 0) {
		DPRINTK("Invalid CPU number=%d\n", param->cpu);
		return -EINVAL;
	}

	spin_lock(&lkst_buffer_lock);

			/* Wait until all the event handlers exit. */
	while (atomic_read(&(lkst.refcnt)));

	head_id = lkst.buffer_head[param->cpu];
	tail_id = lkst.buffer_tail[param->cpu];

	if (!LKST_POINTER_VERIFY_STRICT(lkst.buffer_table[head_id].bufferp, head_buffer)) {
		DPRINTK("Buffer head (ID=%d) of CPU=%d does not exist!\n", head_id, param->cpu);
		spin_unlock(&lkst_buffer_lock);
		return -EBUSY;
	}

	if (!LKST_POINTER_VERIFY_STRICT(lkst.buffer_table[tail_id].bufferp, tail_buffer)) {
		DPRINTK("Buffer tail (ID=%d) of CPU=%d does not exist!\n", tail_id, param->cpu);
		spin_unlock(&lkst_buffer_lock);
		return -EBUSY;
	}

	if (param->ring) {
		/* Change buffer link structure to buffer ring */
		if (head_id == tail_id) {
			DPRINTK("Multiple buffers are needed on the buffer link!\n", "");
			spin_unlock(&lkst_buffer_lock);
			return -EINVAL;
		}

		lkst.buffer_ring[param->cpu] = 1;

		/* Connect head and tail of buffer link */
		head_buffer->prev = tail_id;
		tail_buffer->next = head_id;

	} else {
		/* Change buffer link structure to buffer list */

		lkst.buffer_ring[param->cpu] = 0;

		/* Disconnect head and tail of buffer link */
		head_buffer->prev = LKST_BUFFER_ID_VOID;
		tail_buffer->next = LKST_BUFFER_ID_VOID;
	}

	spin_unlock(&lkst_buffer_lock);

	return (0);
}


int lkst_buffer_setrmod_core(struct lkst_buffer_srmodparam *sp,
			     struct file *filp)
{
	struct lkst_event_buffer *head_buffer;
	lkst_buffer_id head_id;
	int cpu;

	/* !NOTE!: please check (filp != NULL && sp != NULL) before calling this.*/
	if ((lkst_buffer_id) (filp->private_data) != LKST_BUFFER_ID_VOID) {
		DPRINTK("This process(pid=%d) is already set to read other buffer list.\n", current->pid);
		return -EBUSY;
	}

	cpu = sp->cpu;
	
	if (cpu >= smp_num_cpus || cpu < 0) {
		DPRINTK("Invalid CPU number=%d\n", cpu);
		return -EINVAL;
	}

	/* CPU binding */
	lkst_cpu_migrate_set(cpu);

	if (lkst.buffer_owner[cpu] != (PID_MAX + 1)) {
		DPRINTK("Specified buffer list is already set reading mode by pid=%d\n",
			lkst.buffer_owner[cpu]);
		return -EBUSY;
	}

	// spin_lock(&lkst_buffer_lock); CPU migrate

	head_id = lkst.buffer_head[cpu];

	if (!LKST_POINTER_VERIFY_STRICT(lkst.buffer_table[head_id].bufferp, head_buffer)) {
		DPRINTK("Buffer head (ID=%d) of CPU=%d does not exist!\n", head_id, cpu);
		// spin_unlock(&lkst_buffer_lock); CPU migrate
		return -EBUSY;
	}

	if (!lkst.buffer_ring[cpu]) {
		/* buffer is in list structure */
		(lkst_buffer_id) (filp->private_data) = head_id;
	} else {
		/* buffer is in ring structure */
		(lkst_buffer_id) (filp->private_data) =
			lkst.current_buffer_id[cpu];
	}

	/* Set owner pid to the buffer list of specified CPU */ 
	lkst.buffer_owner[cpu] = current->pid;

	/* get buffer time stamp */
	sp->tsc = lkst.tsc;
	sp->xtime.tv_sec = lkst.xtime.tv_sec;
	sp->xtime.tv_usec = lkst.xtime.tv_usec;
	sp->cpu_freq = lkst.cpu_freq;
	sp->buf_ver = lkst.buf_ver;
	strncpy(sp->arch, lkst.arch, LKST_ARCH_NAME_LEN);
	sp->endian_big = lkst.endian_big;

	// spin_unlock(&lkst_buffer_lock); CPU migrate

	return (0);
}


void lkst_buffer_dump_current_txt(void)
{
	int i;
	int cpu;
	lkst_maskset_id id;
	int err_f;
	struct lkst_event_record *buff_body;
	size_t buff_size;
	unsigned int *pArgs;
	lkst_buffer_id buffer_id;
	struct lkst_event_buffer *buffer;
	int	event_type;
	char	*event_type_name;

	/* stop tracing */
	id = LKST_MASKSET_ID_RNOTHING;
	err_f = lkst_maskset_xchg(&id);

	if (!LKST_POINTER_VERIFY_STRICT
	    (lkst.buffer_table[0].bufferp, buffer)) {
		DPRINTK("Kernel State Tracer was not initialized\n", "");
	} else {
		for (cpu = 0; cpu < smp_num_cpus; cpu++) {
			buffer_id =lkst.current_buffer_id[cpu];
			if (!((buffer_id <= LKST_BUFFER_ID_MAX) &&
			      (LKST_POINTER_VERIFY_STRICT(lkst.buffer_table[buffer_id].bufferp, buffer)))) 
				continue;

			buff_size = buffer->size - 1;

			for (i = 0; i <= (buff_size + 1 - sizeof(struct lkst_event_record)); i += sizeof(struct lkst_event_record)) {
				(char *)buff_body = (char *)LKST_BUFFER_BODY_TOP(buffer) +
					((i + buffer->writep.counter) & (unsigned int)buff_size);
				event_type = buff_body->log_event_type;

					/* Stop dump if buffer pointer 
					   reaches unused entry. */
				if (!LKST_ETYPE_VERIFY(event_type)) continue;

				event_type_name = lkst_event_type_descriptions[event_type];

				printk("0x%08x: %016llu [mc], \\\n"
				       " type: %25s(0x%04x), pid: %010u, cpu: %03u \\\n",
				       buff_body->log_recid,
				       buff_body->log_time,
				       event_type_name, event_type,
				       buff_body->log_pid,
				       cpu);

				pArgs = (unsigned int *) &buff_body->log_arg1;
				printk("  0x%08x 0x%08x 0x%08x 0x%08x \\\n"
				       "  0x%08x 0x%08x 0x%08x 0x%08x \\\n\n",
				       pArgs[0], pArgs[1], pArgs[2], pArgs[3], 
				       pArgs[4], pArgs[5], pArgs[6], pArgs[7]);

			}	/* i */
		}		/* cpu */

	}

	/* restart tracing */
	printk("id: %010u\n", id);
	if (!err_f)
		lkst_maskset_xchg(&id);

	return ;
}

/*
 * lkst_evhandler_register
 * return value
 *	success 0
 *	fail	errno
 * Called from external modules
 */
int lkst_evhandler_register(lkst_evhandler_id id, const char *name,
			    lkst_evhandler_func evhandler,
			    lkst_evhandler_ctrl_func evhandler_ctrl)
{
	register int ret;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (evhandler == NULL) {
		DPRINTK("NULL handler(please use lkst_evhandler_unregister, instead)\n");
		return -EINVAL;
	}

	/* spin_lock(lkst.evhandler_table_lock) */
	if (name == NULL || name[0] == '\0' ){
		DPRINTK("Invalid evhandler name!\n");
		/* spin_unlock(lkst.evhandler_table_lock) */
		return -EINVAL;
	}
	if ( lkst_evhandler_get_id_core(name) != LKST_EVHANDLER_ID_VOID) {
		DPRINTK("The name of handler <%s> is already used.\n",name);
		/* spin_unlock(lkst.evhandler_table_lock) */
		return -EINVAL;
	}

	if (id <= LKST_EVHANDLER_ID_RESERVED) { 
		DPRINTK("Reserved handler ID = %d\n", id);
		/* spin_unlock(lkst.evhandler_table_lock) */
		return -EINVAL;
	}
	if (id == LKST_EVHANDLER_ID_VOID) {
		/* Get an unused ID. */
		for (id = LKST_EVHANDLER_ID_RESERVED + 1; 
		     lkst.evhandler_table[id].evhandler != NULL ; id++) 
			if(id > LKST_EVHANDLER_ID_MAX){
				DPRINTK("No handler ID is available.\n");
				/* spin_unlock(lkst.evhandler_table_lock) */
				return -ENOSPC;
			}
	} else if (lkst.evhandler_table[id].evhandler != NULL) {
		return -EINVAL;
	}
	
	ret = lkst_evhandler_register_core(id, name, 
					evhandler, evhandler_ctrl);
	/* spin_unlock(lkst.evhandler_table_lock) */
	return ret == 0 ? id : ret;
}

int lkst_evhandler_unregister(lkst_evhandler_id id)
{
	register int ret;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
	/* spin_lock(lkst.evhandler_table_lock) */
	if (lkst.evhandler_table[id].evhandler == NULL) {
		DPRINTK("Specified handler (ID = %d) does not exist.\n", id);
		return -EINVAL;
	}
	if (id <= LKST_EVHANDLER_ID_RESERVED) { 
		DPRINTK("Reserved handler ID = %d\n", id);
		return -EINVAL;
	}
	ret = lkst_evhandler_unregister_core(id);
	/* spin_unlock(lkst.evhandler_table_lock) */
	return ret;
}

#if DEBUG
int lkst_ioc_diag(long arg)
{
	int cpu = smp_processor_id();

	if (copy_from_user(&(lkst_diag_argdata[cpu]),
		       (struct lkst_diag_arg *) arg,
			   sizeof(lkst_diag_argdata[cpu]))) {
  		DPRINTK("Reading fail at user memory address %p\n", arg);
		return -EINVAL;
	}


	return ((*lkst_diag_argdata[cpu].func) ((char *) &(lkst_diag_argdata[cpu].arg)));
}
#endif /* DEBUG */


/* kernel trc_status wrap function */
int lkst_trc_status(struct lkst_status_param *param)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (param == NULL) { 
		printk(KERN_ERR "Warning : argument param is NULL!\n");
		return -EINVAL;
	}

	return lkst_trc_status_core(param);
}

/* kernel trc_start wrap function */
int lkst_trc_start()
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_trc_start_core();
}

/* kernel trc_stop wrap function */
int lkst_trc_stop()
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_trc_stop_core();
}

/* kernel maskset_read wrap function */
int lkst_maskset_read(struct lkst_maskset_param *param)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (param == NULL) { 
		printk(KERN_ERR "Warning : argument param is NULL!\n");
		return -EINVAL;
	}

	return lkst_maskset_read_core(param);
}

/* kernel maskset_write wrap function */
int lkst_maskset_write(struct lkst_maskset_param *param)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (param == NULL) { 
		printk(KERN_ERR "Warning : argument param is NULL!\n");
		return -EINVAL;
	}

	return lkst_maskset_write_core(param);
}

/* kernel maskset_xchg wrap function */
int lkst_maskset_xchg(lkst_maskset_id * id)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (id == NULL) { 
		printk(KERN_ERR "Warning : argument id is NULL!\n");
		return -EINVAL;
	}

	return lkst_maskset_xchg_core(id);
}

/* kernel maskset_set wrap function */
int lkst_maskset_set(lkst_maskset_id id)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_maskset_xchg_core(&id);
}

/* kernel maskset_list wrap function */
int lkst_maskset_list(struct lkst_maskset_listparam *lp)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (lp == NULL) { 
		printk(KERN_ERR "Warning : argument lp is NULL!\n");
		return -EINVAL;
	}

	return lkst_maskset_list_core(lp);
}

/* kernel maskset_delete wrap function */
int lkst_maskset_delete(lkst_maskset_id id)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_maskset_delete_core(id);
}

/* kernel evhandler_list wrap function */
int lkst_evhandler_list(struct lkst_evhandler_listparam *lp)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (lp == NULL) { 
		printk(KERN_ERR "Warning : argument lp is NULL!\n");
		return -EINVAL;
	}

	return lkst_evhandler_list_core(lp);
}

/* kernel evhandler_ctrl wrap function */
int lkst_evhandler_ctrl(struct lkst_evhandler_ctrl_param
			*evhandler_ctrl_param)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (evhandler_ctrl_param == NULL) { 
		printk(KERN_ERR "Warning : argument evhandler_ctrl_param is NULL!\n");
		return -EINVAL;
	}

	return lkst_evhandler_ctrl_core(evhandler_ctrl_param);
}

/* kernel evhandler_get_id wrap function */
int lkst_evhandler_get_id(const char *name)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;
	
	if ( name == NULL || name[0] == '\0' ) {
		printk(KERN_ERR "Warnig: argument name is NULL!\n");
		return -EINVAL;
	}
	return lkst_evhandler_get_id_core(name);
}

/* kernel buffer_read wrap function */
int lkst_buffer_read(struct lkst_log_buffer *lbuffer)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (lbuffer == NULL) { 
		printk(KERN_ERR "Warning : argument lbuffer is NULL!\n");
		return -EINVAL;
	}

	return lkst_buffer_read_core(lbuffer, 0);
}

#if DEBUG
/* kernel buffer_seek wrap function */
int lkst_buffer_seek(struct lkst_buffer_seekparam *bs)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (bs == NULL) { 
		printk(KERN_ERR "Warning : argument bs is NULL!\n");
		return -EINVAL;
	}

	return lkst_buffer_seek_core(bs);
}
#endif /* DEBUG */

/* kernel buffer_create wrap function */
int lkst_buffer_create(struct lkst_buffer_param *param)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (param == NULL) { 
		printk(KERN_ERR "Warning : argument param is NULL!\n");
		return -EINVAL;
	}
	
	return lkst_buffer_create_core(param);
}

/* kernel buffer_shift wrap function */
int lkst_buffer_shift(int cpu)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_buffer_shift_core(cpu);
}

/* kernel buffer_delete wrap function */
int lkst_buffer_delete(lkst_buffer_id id)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_buffer_delete_core(id);
}

/* kernel buffer_list wrap function */
int lkst_buffer_list(struct lkst_buffer_listparam *lp)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (lp == NULL) { 
		printk(KERN_ERR "Warning : argument lp is NULL!\n");
		return -EINVAL;
	}

	return lkst_buffer_list_core(lp);
}

/* kernel buffer_ring wrap function */
int lkst_buffer_ring(struct lkst_buffer_ringparam *rp)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (rp == NULL) { 
		printk(KERN_ERR "Warning : argument rp is NULL!\n");
		return -EINVAL;
	}
	
	return lkst_buffer_ring_core(rp);
}


/*********************************************************************
 * ioctl functions (static *CLOSED* functions)
 ********************************************************************/

/* ioctl trc_status wrap function */
static int lkst_trc_status_ioctl(struct lkst_status_param *param_user)
{
	int retval = 0;
	struct lkst_status_param param;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	retval = lkst_trc_status_core(&param);

	if (copy_to_user(param_user, &param, sizeof(struct lkst_status_param))) {
		DPRINTK("Writing fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	return retval;
}

/* ioctl trc_start wrap function */
static int lkst_trc_start_ioctl(void)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_trc_start_core();
}

/* ioctl trc_stop wrap function */
static int lkst_trc_stop_ioctl(void)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_trc_stop_core();
}

/* ioctl maskset_read wrap function */
static int lkst_maskset_read_ioctl(struct lkst_maskset_param *param_user)
{
	int retval = 0;
	struct lkst_maskset_param param;
	struct lkst_maskset_body *maskset;
	struct lkst_maskset_body *user_maskset;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user, sizeof(struct lkst_maskset_param))) {
		DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	if (param.maskset_size <= 0 || verify_area(VERIFY_WRITE, param.maskset, param.maskset_size)) {
		DPRINTK("Invalid argument(maskset/maskset_size)\n");
		return -EINVAL;
	}

	user_maskset = param.maskset;
	maskset = vmalloc(param.maskset_size);
	if (maskset == NULL) {
		DPRINTK("Cannot allocate kernel memory : size=%d\n", param.maskset_size);
		return -ENOMEM;
	}

	param.maskset = maskset;
	retval = lkst_maskset_read_core(&param);

	if (copy_to_user(user_maskset, param.maskset, param.maskset_size)) {
		DPRINTK("Writing fail at user memory address %p\n", user_maskset);
		return -EINVAL;
	}

	param.maskset = user_maskset;

	if (copy_to_user(param_user, &param, sizeof(struct lkst_maskset_param))) {
		DPRINTK("Writing fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	vfree(maskset);

	return retval;
}

/* ioctl maskset_write wrap function */
static int lkst_maskset_write_ioctl(struct lkst_maskset_param *param_user)
{
	int retval = 0;
	struct lkst_maskset_param param;
	struct lkst_maskset_body *maskset, *save_user_maskset;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user, sizeof(struct lkst_maskset_param))) {
		DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	if (param.maskset_size <= 0 || verify_area(VERIFY_READ, param.maskset, param.maskset_size)) {
		DPRINTK("Ibvalid argument(maskset/maskset_size)\n", "");
		return -EINVAL;
	}

	maskset = vmalloc(param.maskset_size);
	if (maskset == NULL) {
		DPRINTK("Cannot allocate kernel memory : size=%d\n", param.maskset_size);
		return -ENOMEM;
	}

	if (copy_from_user(maskset, param.maskset, param.maskset_size)) {
		DPRINTK("Reading fail at user memory address %p\n", param.maskset);
		return -EINVAL;
	}

	save_user_maskset = param.maskset;
	param.maskset = maskset;

	retval = lkst_maskset_write_core(&param);

	param.maskset = save_user_maskset;

	if (copy_to_user(param_user, &param, sizeof(struct lkst_maskset_param))) {
		DPRINTK("Writing fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	vfree(maskset);

	return retval;
}

/* ioctl maskset_set wrap function */
static int lkst_maskset_set_ioctl(lkst_maskset_id id)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_maskset_xchg_core(&id);
}

/* ioctl maskset_list wrap function */
static int lkst_maskset_list_ioctl(struct lkst_maskset_listparam *param_user)
{
	int retval = 0;
	struct lkst_maskset_listparam param;
	struct lkst_maskset_listent *listent;
	struct lkst_maskset_listent *user_listent;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user, sizeof(struct lkst_maskset_listparam))) {
		DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}


	if (param.listent_size <= 0 || verify_area(VERIFY_WRITE, param.listent, param.listent_size)) {
		DPRINTK("Invalid argument(listent/listent_size)\n", "");
		return -EINVAL;
	}
	
	user_listent = param.listent;
	listent = vmalloc(param.listent_size);
	if (listent == NULL) {
		DPRINTK("Cannot allocate kernel memory : size=%d\n", param.listent_size);
		return -ENOMEM;
	}

	param.listent = listent;

	retval = lkst_maskset_list_core(&param);

	if (copy_to_user(user_listent, param.listent, param.listent_size)) {
		DPRINTK("Writing fail at user memory address %p\n", user_listent);
		return -EINVAL;
	}

	param.listent = user_listent;

	if (copy_to_user(param_user, &param,
			 sizeof(struct lkst_maskset_listparam))) {
		DPRINTK("Writing fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	vfree(listent);

	return retval;
}

/* ioctl maskset_delete wrap function */
static int lkst_maskset_delete_ioctl(lkst_maskset_id id)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_maskset_delete_core(id);
}

/* ioctl evhandler_list wrap function */
static int lkst_evhandler_list_ioctl(struct lkst_evhandler_listparam *param_user)
{
	int retval = 0;
	struct lkst_evhandler_listparam param;
	struct lkst_evhandler_listent *listent;
	struct lkst_evhandler_listent *user_listent;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user, sizeof(struct lkst_evhandler_listparam))) {
		DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}


	if (param.listent_size <= 0 || verify_area(VERIFY_WRITE, param.listent, param.listent_size)) {
		DPRINTK("Invalid argument(listent/listent_size)\n", "");
		return -EINVAL;
	}
	
	user_listent = param.listent;
	listent = vmalloc(param.listent_size);
	if (listent == NULL) {
		DPRINTK("Cannot allocate kernel memory : size=%d\n", param.listent_size);
		return -ENOMEM;
	}

	param.listent = listent;

	retval = lkst_evhandler_list_core(&param);

	if (copy_to_user(user_listent, param.listent, param.listent_size)) {
		DPRINTK("Writing fail at user memory address %p\n", user_listent);
		return -EINVAL;
	}

	vfree(listent);

	return retval;
}

/* ioctl evhandler_ctrl wrap function */
static int lkst_evhandler_ctrl_ioctl(struct lkst_evhandler_ctrl_param *param_user)
{
	int retval = 0;
	struct lkst_evhandler_ctrl_param param;
	char *buf = NULL;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user, sizeof(struct lkst_evhandler_ctrl_param))) {
		DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}


	if (param.bufsize < 0 || verify_area(VERIFY_READ, param.buf, param.bufsize)) {
		DPRINTK("Invalid argument(buf/bufsize)\n", "");
		return -EINVAL;
	}

	if (param.bufsize != 0) {
		buf = vmalloc(param.bufsize);
		if (buf == NULL) {
			DPRINTK("Cannot allocate kernel memory : size=%d\n", param.bufsize);
			return -ENOMEM;
		}

		if (copy_from_user(buf, param.buf, param.bufsize)) {
			DPRINTK("Reading fail at user memory address %p\n", param.buf);
			return -EINVAL;
		}
	}

	param.buf = buf;

	retval = lkst_evhandler_ctrl_core(&param);

	if (copy_to_user(param_user, &param, sizeof(struct lkst_evhandler_ctrl_param))) {
  		DPRINTK("Writing fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	if (buf != NULL)
		vfree(buf);

	return retval;
}

/* ioctl buffer_read wrap function */
static int lkst_buffer_read_ioctl(struct lkst_log_buffer *lbuffer_user)
{
	int retval = 0;
	struct lkst_log_buffer lbuffer;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&lbuffer, lbuffer_user, sizeof(struct lkst_log_buffer))) {
	  	DPRINTK("Reading fail at user memory address %p\n", lbuffer_user);
		return -EINVAL;
	}


	if (lbuffer.read_size <= 0 || verify_area(VERIFY_WRITE, lbuffer.buffer, lbuffer.read_size)) {
		DPRINTK("Invalid argument(buffer/read_size)\n", "");
		return -EINVAL;
	}

	retval = lkst_buffer_read_core(&lbuffer, 1);

	if (copy_to_user(lbuffer_user, &lbuffer, sizeof(struct lkst_log_buffer))) {
		DPRINTK("Writing fail at user memory address %p\n", lbuffer_user);
		return -EINVAL;
	}

	return retval;
}

#if DEBUG
/* ioctl buffer_seek wrap function */
static int lkst_buffer_seek_ioctl(struct lkst_buffer_seekparam *bs_user)
{
	int retval = 0;
	struct lkst_buffer_seekparam bs;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&bs, bs_user, sizeof(struct lkst_buffer_seekparam))) {
		DPRINTK("Reading fail at user memory address %p\n", bs_user);
		return -EINVAL;
	}

	retval = lkst_buffer_seek_core(&bs);

	if (copy_to_user(bs_user, &bs, sizeof(struct lkst_buffer_seekparam))) {
		DPRINTK("Writing fail at user memory address %p\n", bs_user);
		return -EINVAL;
	}

	return retval;
}
#endif /* DEBUG */

/* ioctl buffer_create wrap function */
static int lkst_buffer_create_ioctl(struct lkst_buffer_param *param_user)
{
	int retval = 0;
	struct lkst_buffer_param param;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user, sizeof(struct lkst_buffer_param))) {
		DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	retval = lkst_buffer_create_core(&param);

	if (copy_to_user(param_user, &param, sizeof(struct lkst_buffer_param))) {
		DPRINTK("Writing fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	return retval;
}


/* ioctl buffer_shift wrap function */
static int lkst_buffer_shift_ioctl(int arg)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_buffer_shift_core(arg);
}

/* ioctl buffer_delete wrap function */
static int lkst_buffer_delete_ioctl(lkst_buffer_id arg)
{
	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	return lkst_buffer_delete_core(arg);
}

/* ioctl buffer_list wrap function */
static int lkst_buffer_list_ioctl(struct lkst_buffer_listparam *param_user)
{
	int retval = 0;
	struct lkst_buffer_listparam param;
	struct lkst_buffer_listent *listent;
	struct lkst_buffer_listent *user_listent;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user, sizeof(struct lkst_buffer_listparam))) {
		DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}


	if (param.listent_size <= 0 || verify_area(VERIFY_WRITE, param.listent, param.listent_size)) {
		DPRINTK("Invalid argument(listent/listent_size)\n", "");
		return -EINVAL;
	}
	
	user_listent = param.listent;
	listent = vmalloc(param.listent_size);
	if (listent == NULL) {
		DPRINTK("Cannot allocate kernel memory : size=%d\n", param.listent_size);
		return -ENOMEM;
	}

	param.listent = listent;

	retval = lkst_buffer_list_core(&param);

	if (copy_to_user(user_listent, param.listent, param.listent_size)) {
		DPRINTK("Writing fail at user memory address %p\n", user_listent);
		return -EINVAL;
	}
	vfree(listent);

	return retval;
}

/* ioctl buffer_ring wrap function */
static int lkst_buffer_ring_ioctl(struct lkst_buffer_ringparam *param_user)
{
	struct lkst_buffer_ringparam param;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user, sizeof(struct lkst_buffer_ringparam))) {
		DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	return lkst_buffer_ring_core(&param);
}

/* ioctl buffer_setrmod wrap function */
static int lkst_buffer_setrmod_ioctl(struct lkst_buffer_srmodparam *param_user,
			  struct file *filp)
{
	int retval = 0;
	struct lkst_buffer_srmodparam param;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&param, param_user, sizeof(struct lkst_buffer_srmodparam))) {
		DPRINTK("Reading fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	retval = lkst_buffer_setrmod_core(&param, filp);

	if (copy_to_user(param_user, &param, sizeof(struct lkst_buffer_srmodparam))) {
		DPRINTK("Writing fail at user memory address %p\n", param_user);
		return -EINVAL;
	}

	return retval;
}

/* ioctl entry_log wrap function */
static int lkst_entry_log_ioctl(struct lkst_entry_args *arg_user)
{
	struct lkst_entry_args arg;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (copy_from_user(&arg, arg_user, sizeof(struct lkst_entry_args))) {
		DPRINTK("Reading fail at user memory address %p\n", arg_user);
		return -EINVAL;
	}

	(void) lkst_entry_log(NULL, arg.event_type,
			      arg.arg1, arg.arg2, arg.arg3, arg.arg4);

	return (0);
}


/* LKST device open function */
int lkst_open(struct inode *inode, struct file *filp)
{

#ifdef MODULE
	/* Increment module reference count */
	MOD_INC_USE_COUNT;
#endif /* MODULE */

	/* Initialize reading ID of buffer which used by "lkstlogd" */
	(lkst_buffer_id) filp->private_data = LKST_BUFFER_ID_VOID;

	return (0);
}


/* LKST device close function */
int lkst_close(struct inode *inode, struct file *filp)
{

	int i;
#ifdef MODULE
	/* Decrement module reference count */
	MOD_DEC_USE_COUNT;
#endif /* MODULE */

	/* Clear owner pid of buffer list */
	for (i = 0; i < LKST_CPU_MAX; i++) {
		if (lkst.buffer_owner[i] == current->pid) {
			lkst.buffer_owner[i] = PID_MAX + 1;
			break;
		}
	}

	/* CPU binding */
	lkst_cpu_migrate_reset();

	return (0);
}


/* LKST device read function */
ssize_t lkst_read(struct file *filp, char *buf, size_t count, loff_t *ppos)
{
	ssize_t retval, result_size = 0;
	lkst_buffer_id id;
	struct lkst_log_buffer lbuffer;
	struct lkst_event_buffer *buffer;
	int cpu;
	int rcnt, wcnt, rwdiff;

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if ((lkst_buffer_id) (filp->private_data) == LKST_BUFFER_ID_VOID) {
		DPRINTK("This process(pid=%d) is not initialized in LKST, Please use command ioctl(%d,..)\n", current->pid, LKST_IOC_BUFFER_SETRMOD);
		return -EBUSY;
	}

	if (count <= 0 || verify_area(VERIFY_WRITE, buf, count)) {
		DPRINTK("Invalid argument(buf/count)\n", "");
		return -EINVAL;
	}

	id = (lkst_buffer_id) (filp->private_data);

	do {
		if (!LKST_POINTER_VERIFY(lkst.buffer_table[id].bufferp, buffer)) {
			DPRINTK("Buffer(ID=%d) does not exist!\n", id);
			return -EBUSY;
		}

		cpu = buffer->cpu;

		rcnt = atomic_read(&(buffer->rcnt));

		/* Increment buffer read count */
		if ((rcnt % 2) == 0) {
			atomic_inc(&(buffer->rcnt));
			rcnt++;
		}

		wcnt = atomic_read(&(buffer->wcnt));

		rwdiff = wcnt - rcnt;

		if (rwdiff >= 2 || rwdiff < 0) {
			if (rwdiff <= -2) /* wcnt/rcnt overflow */
				atomic_set(&(buffer->rcnt),
					   wcnt - 1 + (rwdiff % 2));
			else if (rwdiff >= 2) /* buffer is wrapped */
				atomic_set(&(buffer->rcnt),
					   wcnt - 1 - (rwdiff % 2));
			LKST_BUFFER_READP_RESET(buffer);
			id = buffer->next;
			/* Set last read buffer ID */
			(lkst_buffer_id) (filp->private_data) = id;
			return -EAGAIN;
		} else if (rwdiff == 0) {
			/* process sleep on */
			lkst.owner_ready[cpu]=0;
			interruptible_sleep_on(&lkst.buffer_wq[cpu]);
		}

		lbuffer.read_size = count;
		(char *) lbuffer.buffer = buf + result_size;
		lbuffer.id = id;
		lbuffer.result_read_size = 0;

		retval = lkst_buffer_read_core(&lbuffer, 2);

		if (retval)
			return retval;

		if (signal_pending(current)) /* singal pending */
			return lbuffer.result_read_size + result_size;

		/* Read whole buffer */
		if (lbuffer.id != id) {
			LKST_BUFFER_READP_RESET(buffer);
			id = lbuffer.id;
			atomic_inc(&(buffer->rcnt));
			/* Set last read buffer ID */
			(lkst_buffer_id) (filp->private_data) = id;
		}

		count -= lbuffer.result_read_size;

		result_size += lbuffer.result_read_size;

	} while (count >= LKST_BUFFER_SIZE_LOCAL2EXT(LKST_BUFFER_HEADSPACE));

	return result_size;
}


/* LKST device ioctl function */
int lkst_ioctl(struct inode *inode, struct file *filp,
	       unsigned int request, unsigned long arg)
{
	int retval = 0;		/* return value */

	switch (request) {
	case LKST_IOC_TRC_STATUS:
		retval = 
		    lkst_trc_status_ioctl((struct lkst_status_param *) arg);
		break;
	case LKST_IOC_TRC_START:
		retval = lkst_trc_start_ioctl();
		break;
	case LKST_IOC_TRC_STOP:
		retval = lkst_trc_stop_ioctl();
		break;
	case LKST_IOC_MASKSET_READ:
		retval =
		    lkst_maskset_read_ioctl((struct lkst_maskset_param *) arg);
		break;
	case LKST_IOC_MASKSET_WRITE:
		retval =
		   lkst_maskset_write_ioctl((struct lkst_maskset_param *) arg);
		break;
	case LKST_IOC_MASKSET_SET:
		retval = lkst_maskset_set_ioctl((lkst_maskset_id) arg);
		break;
	case LKST_IOC_MASKSET_LIST:
		retval =
		    lkst_maskset_list_ioctl((struct lkst_maskset_listparam *) arg);
		break;
	case LKST_IOC_MASKSET_DELETE:
		retval = lkst_maskset_delete_ioctl((lkst_maskset_id) arg);
		break;
	case LKST_IOC_EVHANDLER_LIST:
		retval =
		    lkst_evhandler_list_ioctl((struct lkst_evhandler_listparam *) arg);
		break;
	case LKST_IOC_EVHANDLER_CTRL:
		retval =
		    lkst_evhandler_ctrl_ioctl((struct lkst_evhandler_ctrl_param *) arg);
		break;

	case LKST_IOC_BUFFER_READ:
		retval = 
		    lkst_buffer_read_ioctl((struct lkst_log_buffer *) arg);
		break;
#if DEBUG
	case LKST_IOC_BUFFER_SEEK:
		retval = 
		    lkst_buffer_seek_ioctl((struct lkst_buffer_seekparam *) arg);
		break;
#endif /* DEBUG */
	case LKST_IOC_BUFFER_CREATE:
		retval =
		    lkst_buffer_create_ioctl((struct lkst_buffer_param *) arg);
		break;
	case LKST_IOC_BUFFER_SHIFT:
		retval = lkst_buffer_shift_ioctl((int) arg);
		break;
	case LKST_IOC_BUFFER_DELETE:
		retval = lkst_buffer_delete_ioctl((lkst_buffer_id) arg);
		break;

	case LKST_IOC_BUFFER_LIST:
		retval = lkst_buffer_list_ioctl((struct lkst_buffer_listparam *) arg);
		break;

	case LKST_IOC_BUFFER_RING:
		retval = lkst_buffer_ring_ioctl((struct lkst_buffer_ringparam *) arg);
		break;

	case LKST_IOC_BUFFER_SETRMOD:
		retval = lkst_buffer_setrmod_ioctl((struct lkst_buffer_srmodparam *) arg, filp);
		break;

	case LKST_IOC_ENTRY_LOG:
		retval =
		    lkst_entry_log_ioctl((struct lkst_entry_args *) arg);
		break;
	default:
		DPRINTK("Invalid command no.%d\n", request);
		return -EINVAL;
#if DEBUG
	case LKST_IOC_DIAG:
		retval = lkst_ioc_diag(arg);
		break;
#endif /* DEBUG */
	}

	return retval;
}

/*********************************************************************
 * Initialize/cleanup functions
 *********************************************************************/
/*
 * function lkst_trc_init_stage0
 * Initialize the first part of Kernel State Tracer.
 * This function initializes data not concerned with SMP.
 * 
 *  Arguments
 *    NONE
 *  returns
 *    NONE
 * 
 */
void lkst_trc_init_stage0()
{
	int i;

	/* The following values are initialized when kernel was loaded,
	 *   lkst.cpu_recid[...] = 1;
	 *   maskset_rnothing[], maskset_rall[], maskset_rdefault[]
	 */

	printk("LKST: Starting initialization stage0\n");

#ifdef CONFIG_MAGIC_SYSRQ
	register_sysrq_key('v', &sysrq_lkst_dump_buffer_op);
#endif /* CONFIG_MAGIC_SYSRQ */

	/* Initialize LKST header time stamp */
	lkst_time_set(0);

	/* Initialize stage0 buffer */
	lkst_stage0_event_buffer.cmos_time_flag = lkst.cmos_time_flag;
	lkst_stage0_event_buffer.xtime.tv_sec = lkst.xtime.tv_sec;
	lkst_stage0_event_buffer.xtime.tv_usec = lkst.xtime.tv_usec;

	/* set buffer_table */
	lkst.buffer_table[0].bufferp = &lkst_stage0_event_buffer;
	LKST_POINTER_CHECK_INIT(lkst.buffer_table[0].bufferp);

	/* set current buffer */
	lkst.current_buffer_id[0] = 0;

	/* initialize current_buffer */
	for (i = 1; i < LKST_CPU_MAX; i++) {
		lkst.current_buffer_id[i] = LKST_BUFFER_ID_VOID;
	}

	/* initialize buffer_head/tail */
	for (i = 0; i < LKST_CPU_MAX; i++) {
		lkst.buffer_head[i] = LKST_BUFFER_ID_VOID;
		lkst.buffer_tail[i] = LKST_BUFFER_ID_VOID;
		lkst.buffer_ring[i] = 0;
		lkst.buffer_owner[i] = PID_MAX + 1;
		lkst.buffer_wq[i] = (wait_queue_head_t) __WAIT_QUEUE_HEAD_INITIALIZER(lkst.buffer_wq[i]);
	}

#if DEBUG
	if (LKST_SIZEOF_LKST_EVENT_RECORD == sizeof(struct lkst_event_record)) {
		/* Verification of LKST_SIZEOF_LKST_EVENT_RECORD is failed. */
		printk(KERN_ERR "LKST error(init): sizeof(struct lkst_event_record) is not equal to LKST_SIZEOF_LKST_EVENT_RECORD.\n");
		return;
	}
#endif /* DEBUG */

	/* Clear entries for other Event Handlers */
	for (i = 0; i < LKST_EVHANDLER_TBL_MAX; i++) {
		lkst_evhandler_unregister_core(i);
	}

	/* Initialize Default Event Handler */
	lkst_evhandler_register_core(LKST_EVHANDLER_ID_DEFAULT,
				LKST_EVHANDLER_NAME_DEFAULT,
				lkst_evhandler_default,
				lkst_evhandler_default_ctrl);

	/* Initialize Buffer Shift Event Handler */
	lkst_evhandler_register_core(LKST_EVHANDLER_ID_BUFFER_SHIFT,
				LKST_EVHANDLER_NAME_BUFFER_SHIFT,
				lkst_evhandler_buffer_shift,
				NULL);

	/* Initialize Masetset Exchange Event Handler */
	lkst_evhandler_register_core(LKST_EVHANDLER_ID_MASKSET_XCHG,
				LKST_EVHANDLER_NAME_MASKSET_XCHG,
				lkst_evhandler_maskset_xchg,
				NULL);

#if DEBUG
	/* Initialize Exception Event Handler */
	lkst_evhandler_register_core(LKST_EVHANDLER_ID_EXCEPTION,
				LKST_EVHANDLER_NAME_EXCEPTION,
				lkst_evhandler_exception,
				NULL);
#endif /* DEBUG */

	/* Initialize Shifting buffer And Awaking Daemon Event Handler */
	lkst_evhandler_register_core(LKST_EVHANDLER_ID_BUFFER_SHIFT_DW,
				LKST_EVHANDLER_NAME_BUFFER_SHIFT_DW,
				lkst_evhandler_buffer_shift_dw,
				NULL);


	/* Setup Special Masksets */
	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
				     LKST_MASKSET_ID_RALL,
				     LKST_MASKSET_NAME_RALL, maskset_rall);

	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
				     LKST_MASKSET_ID_RNOTHING,
				     LKST_MASKSET_NAME_RNOTHING,
				     maskset_rnothing);

	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
				     LKST_MASKSET_ID_RDEFAULT,
				     LKST_MASKSET_NAME_RDEFAULT,
				     maskset_rdefault);

	/* Clear entries for other masksets */
	for (i = 3; i < LKST_MASKSET_TBL_MAX; i++) {
		lkst.maskset_table[i].masksetp = NULL;
		lkst.maskset_table[i].masksetp_check = 0;
	}

	/* Setup current maskset pointer */
	LKST_MASKSET_SET_CURRENT(LKST_MASKSET_ID_RDEFAULT);

	/* GKHI hook Initialization */
#ifndef MODULE
	if(hook_init_in_start()){
		printk("LKST: hook_init error\n");
	}
#endif /* MODULE */
	lkst_trc_hook_init();
	lkst_hook_maskset_xchg(lkst.current_maskset_id, LKST_MASKSET_ID_RNOTHING);

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_INIT, LKST_ARG32(0x0001, 0), 0, 0, 0);

	printk("LKST: Initialization stage0 finished\n");

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_INIT, LKST_ARG32(0x00ff, 0), 0, 0, 0);
}


/*
 * function lkst_trc_init_stage1
 * Initialize the second part of Kernel State Tracer.
 * This function initializes buffers, pointers and
 * data concernd with SMP. Therefore, this should be
 * called after all of the CPUs are initialized.
 * 
 *  Arguments
 *    NONE
 *  returns
 *    NONE
 * 
 */
void lkst_trc_init_stage1()
{
	int i;
	struct lkst_buffer_param param;
	struct lkst_event_buffer *buffer;

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_INIT, LKST_ARG32(0x0101, 0), 0, 0, 0);

	printk("LKST: Starting initialization stage1\n");

	/* Set architecture name */
	strncpy(lkst.arch, system_utsname.machine, 65); 

	/* Set cpu frequency */
	lkst.cpu_freq = cpu_khz;

#ifdef	CONFIG_DEBUG_KERNEL_STATE_TRACE_IN_START
	/* Re-initialize time stamp */
	lkst_time_set(0);
	lkst_stage0_event_buffer.cmos_time_flag = lkst.cmos_time_flag;
	lkst_stage0_event_buffer.xtime.tv_sec = lkst.xtime.tv_sec;
	lkst_stage0_event_buffer.xtime.tv_usec = lkst.xtime.tv_usec;
#endif /* CONFIG_DEBUG_KERNEL_STATE_TRACE_IN_START */

	/* initialize buffer_table */
	for (i = 1; i < LKST_BUFFER_TBL_MAX; i++) {
		lkst.buffer_table[i].bufferp = NULL;
		lkst.buffer_table[i].bufferp_check = 0;
	}

	/* allocate event buffer for all CPUs */
	for (i = 0; i < smp_num_cpus; i++) {

		printk("LKST: event buffer allocation CPU=%d\n", i);

		param.id = i + 1;
		param.cpu = i;
		param.size = lkst_init_buffer_size;

		/* create buffer for all CPUs */
		if (lkst_buffer_create(&param)) {
			printk(KERN_ERR "LKST warning(init): Cannot allocate buffer area!\n");
			/* on error, free all buffers except CPU0's. */
			for (i--; i >= 1; i--) {
				vfree((void *)lkst.buffer_table[i + 1].bufferp);
				lkst.buffer_table[i + 1].bufferp = NULL;
				lkst.buffer_table[i + 1].bufferp_check = 0;
			}
			break;
		}

		if (!LKST_POINTER_VERIFY(lkst.buffer_table[i + 1].bufferp,
					 buffer)) {
			printk(KERN_ERR "LKST warning(init): allocated buffer area is collapsed! Please initialize again\n");
			break;
		}

		/* set buffer_head */
		lkst.buffer_head[i] = (i + 1);

		/* increment buffer write count */
		atomic_inc(&(buffer->wcnt));

		/* set current buffer */
		lkst.current_buffer_id[i] = (i + 1);
	}

	// TODO: device special file must be created automatically.
	/* register LKST device file_operations */
#ifdef CONFIG_DEVFS_FS
	if (devfs_register_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME, &lkst_fops))
		printk(KERN_ERR "LKST warning(init):cannot register LKST *conventional* device!\n");
#else
	if (register_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME, &lkst_fops))
		printk(KERN_ERR "LKST error(init):cannot register LKST device!\n");
#endif

#ifdef CONFIG_DEVFS_FS
        lkst_devfs_handle = devfs_register (NULL, LKST_DEV_NAME, DEVFS_FL_DEFAULT, LKST_DEV_MAJOR,
				   LKST_DEV_MINOR, 0020644, &lkst_fops, NULL);
	if ( lkst_devfs_handle == NULL )
		printk(KERN_ERR "LKST error(init):cannot register LKST devfs device!\n");
#endif

	/* LKST initialize flag on */
	lkst_init_flag = LKST_FLAG_YES;

	printk("LKST: initialization stage1 finished\n");

	/* LKST internal event */
	lkst_entry_log(NULL, LKST_ETYPE_LKST_INIT, LKST_ARG32(0x10ff, 0), 0, 0, 0);
}


/*
 * function lkst_trc_init_main
 * Initialize basic part of Kernel State Tracer.
 * This function is called from /init/main.c. and lkst_init_module
 */
void lkst_trc_init_main()
{
#ifndef	CONFIG_DEBUG_KERNEL_STATE_TRACE_IN_START
	lkst_trc_init_stage0();
#endif /* !CONFIG_DEBUG_KERNEL_STATE_TRACE_IN_START */
	lkst_trc_init_stage1();
}

#ifdef MODULE
static int __init lkst_init_module(void)
{
	size_t size;

	printk("LKST: start LKST module.\n");

	/* buffer size check */
	printk("LKST: initial event buffer size=%d\n", lkst_init_buffer_size);
	if (lkst_init_buffer_size < LKST_BUFFER_SIZE_MIN) {
		printk(KERN_ERR "LKST warning(init): buffer size(=%d) is too small, changed to %d\n", lkst_init_buffer_size, LKST_BUFFER_SIZE_MIN);
		lkst_init_buffer_size = LKST_BUFFER_SIZE_MIN;
	} else if (lkst_init_buffer_size > LKST_BUFFER_SIZE_MAX) {
		printk(KERN_ERR "LKST warning(init): buffer size(=%d) is too large, changed to %d\n", lkst_init_buffer_size, LKST_BUFFER_SIZE_MAX);
		lkst_init_buffer_size = LKST_BUFFER_SIZE_MAX;
	}

	for (size = LKST_BUFFER_SIZE_MAX; size > lkst_init_buffer_size; size = size >> 1);
	if (lkst_init_buffer_size != size)
		printk(KERN_ERR "LKST warning(init): buffer size must be power of 2, changed to %d\n", size);
	lkst_init_buffer_size = size;

	lkst_trc_init_main();
	return 0;
}

static void __exit lkst_cleanup_module(void)
{
	int i;

	/* GKHI hook cleanup */
	lkst_trc_hook_cleanup();

	/* LKST initialize flag off */
	lkst_init_flag = LKST_FLAG_NO;

#ifdef CONFIG_MAGIC_SYSRQ
	unregister_sysrq_key('v', &sysrq_lkst_dump_buffer_op);
#endif /* CONFIG_MAGIC_SYSRQ */

	lkst.current_maskset_id = LKST_MASKSET_ID_RNOTHING;
	lkst.current_masksetp = 0;

	memset(lkst.cpu_recid, 1, sizeof(lkst_cpu_recid)*LKST_CPU_RECID_BUF_SIZE);

	/* Cleanup default maskset pointer */
	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
				     LKST_MASKSET_ID_RALL, "", NULL);
	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
				     LKST_MASKSET_ID_RNOTHING, "", NULL);
	LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table,
				     LKST_MASKSET_ID_RDEFAULT, "", NULL);

	/* Cleanup additional maskset pointer & Free */
	for (i = 3; i < LKST_MASKSET_TBL_MAX; i++) {
		if (lkst.maskset_table[i].masksetp) {
			vfree(lkst.maskset_table[i].masksetp);
			LKST_MASKSET_TABLE_ENTRY_SET(lkst.maskset_table, i,
						     "", NULL);
		}
	}

	/* Cleanup event handler */
	lkst_evhandler_unregister_core(LKST_EVHANDLER_ID_DEFAULT);
	for (i = 0; i < LKST_EVHANDLER_TBL_MAX; i++) {
		lkst_evhandler_unregister_core(i);
	}

	/* Cleanup event buffer_table & Deallcation event buffer */
	lkst.buffer_table[0].bufferp = NULL;
	for (i = 1; i < LKST_BUFFER_TBL_MAX; i++) {
		if (lkst.buffer_table[i].bufferp) {
			vfree(lkst.buffer_table[i].bufferp);
			lkst.buffer_table[i].bufferp = NULL;
		}
	}

	/* Cleanup current buffer */
	for (i = 0; i < LKST_CPU_MAX; i++) {
		lkst.buffer_head[i] = LKST_BUFFER_ID_VOID;
		lkst.buffer_tail[i] = LKST_BUFFER_ID_VOID;
		lkst.buffer_ring[i] = 0;
	}

	/* unregister LKST device file_operations */
#ifdef CONFIG_DEVFS_FS
	if (devfs_unregister_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME))
#else
	if (unregister_chrdev(LKST_DEV_MAJOR, LKST_DEV_NAME))
#endif
		printk(KERN_ERR "LKST error :cannot unregister LKST device!\n");

#ifdef CONFIG_DEVFS_FS
        devfs_unregister ( lkst_devfs_handle );
#endif
	/* deallocation complete */
}

module_init(lkst_init_module);
module_exit(lkst_cleanup_module);
MODULE_AUTHOR("LKST project");
MODULE_DESCRIPTION("Linux Kernel State Trace driver");
MODULE_LICENSE("GPL");
MODULE_PARM(lkst_init_buffer_size, "i");
MODULE_PARM_DESC(lkst_init_buffer_size, "Default size of each buffer(4096-0x200000, in power of 2)");
#endif /* MODULE */
