/* $Id: lkst_maskset.h,v 1.3 2002/03/25 19:05:34 serizawa Exp $ */
/* linux/include/linux/lkst_maskset.h */

/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2001-2002 ALL RIGHTS RESERVED.
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by T.Nakamura <tetsu@crl.hitachi.co.jp>
 *
 ************************************************************/

#ifndef _LINUX_LKST_MASKSET_H
#define _LINUX_LKST_MASKSET_H

#include<linux/lkst.h>
#include<linux/lkst_evhandler.h>

#define LKST_MASKSET_NAME_LEN	256

/* LKST_MASKSET_TBL_LEN_MAX means n of entries in a maskset. (cf. LKST_MASKSET_TBL_MAX */
#define LKST_MASKSET_TABLE_LEN_MAX	((LKST_ETYPE_MAX)+1)

#define LKST_MASKSET_NAME_RALL		"RALL"
#define LKST_MASKSET_NAME_RNOTHING	"RNOTHING"
#define LKST_MASKSET_NAME_RDEFAULT	"RDEFAULT"

typedef int lkst_maskset_table_len;

struct lkst_maskset_entry {
	int event_type;			/* corresponding type of evnet */
	lkst_evhandler_id id;	/* event handler ID */
};

struct lkst_maskset_body {
	char name[LKST_MASKSET_NAME_LEN];	/* maskset name */
	lkst_maskset_table_len len;	/* total number of maskset entries */
	struct lkst_maskset_entry entry[LKST_MASKSET_TABLE_LEN_MAX];
					/* maskset entry */
};

struct lkst_maskset_param {
	lkst_maskset_id id;	/* maskset ID */
	size_t maskset_size;	/* maskset size */
	struct lkst_maskset_body *maskset;
				/* address of a maskset stored area */
};

struct lkst_maskset_listent {
	lkst_maskset_id id;			/* maskset ID */
	char name[LKST_MASKSET_NAME_LEN];	/* maskset name */
	lkst_maskset_table_len len;	/* total number of maskset entries */
};

struct lkst_maskset_listparam {
	lkst_maskset_id current_id;	/* current maskset ID */
	size_t listent_size;	/* size of the listent */
	struct lkst_maskset_listent *listent;
				/* area to store the list of masksets */
};


extern int lkst_maskset_read(struct lkst_maskset_param *maskset_param);

extern int lkst_maskset_write(struct lkst_maskset_param *maskset_param);

extern int lkst_maskset_set(lkst_maskset_id id);

extern int lkst_maskset_delete(lkst_maskset_id id);

extern int lkst_maskset_list(struct lkst_maskset_listparam *lp);

#define LKST_MASKSET_SIZE(n)						\
	(sizeof(char)*LKST_MASKSET_NAME_LEN +				\
	     sizeof(lkst_maskset_table_len) +				\
	     sizeof(struct lkst_maskset_entry) * (n))

#define LKST_MASKSET_LISTENT_SIZE(n)					\
	     (sizeof(struct lkst_maskset_listent) * (n))


#endif				/* _LINUX_LKST_MASKSET_H */
