/* $Id: lkst_kernel.c,v 1.2 2002/06/28 11:27:56 indoh Exp $ */
/* linux/kernel/lkst.c */

/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2001-2002 ALL RIGHTS RESERVED.
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by K.SERIZAWA <serizawa@sdl.hitachi.co.jp>
 *
 ************************************************************/

#include <linux/module.h>
#include <linux/lkst_private.h>
#include <linux/vmalloc.h>
#include <linux/init.h>
#include <linux/highmem.h>

EXPORT_SYMBOL(lkst_entry_log);

#ifdef CONFIG_DEBUG_KERNEL_STATE_TRACE
#undef	DEBUG
#endif

#define LKST_DEBUG_LOG(ty, a1, a2, a3, a4)

#define MASKSET_TBL_INIT {"", NULL, (lkst_pointer_check_digit) 0}
#define EVHANDLER_TBL_INIT {"", NULL, (lkst_pointer_check_digit) 0, NULL}
#define CUR_BUF_INIT {NULL, (lkst_pointer_check_digit) 0, 0,(lkst_pointer_check_digit) 0}

/* LKST availabe memory size -- max vmalloc size */
//#define LKST_AVAILABLE_MEMORY_SIZE	(unsigned long)((unsigned long)VMALLOC_END - (unsigned long)VMALLOC_START - (unsigned long)PAGE_SIZE)
#define LKST_AVAILABLE_MEMORY_SIZE	(64<<20)

struct lkst_header lkst = {
	eyecacher:		{'L','K','S','T'},
	refcnt:			ATOMIC_INIT(0),
	cpu_recid:		{[0 ... (LKST_CPU_RECID_BUF_SIZE-1)] = 1},
	current_masksetp:	NULL,
	current_masksetp_check:	(lkst_pointer_check_digit) 0,
	current_maskset_id:	0,
	maskset_table: 		{MASKSET_TBL_INIT, 
				 MASKSET_TBL_INIT,
				 MASKSET_TBL_INIT},
	evhandler_table:	{[0 ... LKST_EVHANDLER_ID_MAX] = EVHANDLER_TBL_INIT},
	current_buffer_id:	{[0 ... (LKST_CPU_MAX - 1)] = LKST_BUFFER_ID_VOID},
	buffer_table:		{{NULL, (lkst_pointer_check_digit) 0}},
	buffer_tail:		{0},
	buffer_head:		{0},
	buffer_ring:		{0},
	buffer_owner: 		{0},
	owner_ready: 		{0},
	buffer_wq:		{{{0}, {0}}},
	maskset_id_saved:	LKST_MASKSET_ID_RDEFAULT,
	status_flag:		ATOMIC_INIT(0), /*obsolate , not used*/
	available_memory_size:	0,
	limit_addr:		NULL,
	endian_big:		LKST_BIG_ENDIAN, /* 1 means BIG_ENDIAN. */
	cpu_freq:		0,
	tsc:			0,
	xtime:			{0 ,0},
	cmos_time_flag:		1,
	buf_ver:		LKST_BUF_VER,
	arch:			{UTS_MACHINE},
};

void lkst_entry_log(void *phookrec, int event_type, lkst_arg_t arg1, lkst_arg_t arg2, 
					lkst_arg_t arg3, lkst_arg_t arg4)
{
	lkst_evhandler_id evhandler_id;
	volatile lkst_maskset *maskset;
	lkst_evhandler_func evhandler;

	if (!LKST_ETYPE_VERIFY(event_type))
		return;

	atomic_inc(&(lkst.refcnt));

#if DEBUG
	{
		int i = LKST_EVHANDLER_MASKSET_TRIAL_MAX;

		/* spin while another cpu may change maskset. */
		while (!LKST_POINTER_VERIFY_STRICT
		       (lkst.current_masksetp, maskset)) {
			if (!--i) {
				LKST_DEBUG_LOG(event_type + 0x200, arg1,
					(unsigned int) (lkst.current_masksetp),
					(unsigned int) (lkst.current_masksetp_check),
					(unsigned int) 0);
				goto out;
			}
		}
	}
#else /* !DEBUG */
	if (!LKST_POINTER_VERIFY_STRICT(lkst.current_masksetp, maskset)) {
		LKST_DEBUG_LOG(event_type + 0x200, arg1,
			       (unsigned int) (lkst.current_masksetp),
			       (unsigned int) (lkst.current_masksetp_check),
			       (unsigned int) 0);
		goto out;
	}
#endif /* DEBUG */

	evhandler_id = LKST_MASKSET_QUERY_EVHANDLER_ID(maskset, event_type);
	if (evhandler_id > LKST_EVHANDLER_ID_MAX)
		goto out;

	if (!LKST_POINTER_VERIFY
	    (lkst.evhandler_table[evhandler_id].evhandler, evhandler)) {
		LKST_DEBUG_LOG(event_type + 0x400, evhandler_id,
			       (unsigned int) (evhandler),
			       (unsigned int) (lkst.evhandler_table[evhandler_id].evhandler_check),
			       (unsigned int) 0);
		goto out;
	}

			/* Avoid infinate loop. */
	if (evhandler == lkst_entry_log)
		goto out;

	(*evhandler) (NULL, event_type, arg1, arg2, arg3, arg4);

out:
	atomic_dec(&(lkst.refcnt));
	return;
}

/*
 * function lkst_time_set
 * This function sets time stamp of LKST header.
 * 
 *  Arguments
 *    from_settimeofday : flag shows to be called from do_sys_settimeofday().
 *  returns
 *    NONE
 * 
 */
void lkst_time_set(int from_settimeofday)
{
	unsigned long flags;

	LKST_CLI(flags);
	if (from_settimeofday)
		lkst.cmos_time_flag = 0;
	get_fast_time(&lkst.xtime);
	lkst.tsc = lkst_evhandlerprim_mc();
	LKST_STI(flags);
}
EXPORT_SYMBOL(lkst_time_set);


/*
 * This function is called by lkst modules to get
 * variable "cpu_khz" from LKST_GET_CPU_KHZ().
 * In some arch, LKST will not provide this function.
 */
#ifdef	LKST_GET_CPU_KHZ
lkst_cpu_freq_t lkst_get_cpu_khz(void)
{
	return LKST_GET_CPU_KHZ();
}
EXPORT_SYMBOL(lkst_get_cpu_khz);
#endif /* ifdef	LKST_GET_CPU_KHZ */


#ifdef CONFIG_DEBUG_KERNEL_STATE_TRACE_MODULE
struct lkst_header *get_lkst(void)
{
	return &lkst;
}
EXPORT_SYMBOL(get_lkst);
#endif /* CONFIG_DEBUG_KERNEL_STATE_TRACE_MODULE*/

// TODO
#ifdef CONFIG_MODULES
/*
 * function lkst_find_module_by_address
 * Find module in which specified address is contained, and return
 * pointer to the module.
 * 
 *  Caution: This function does not verify value of event_entry_p.
 * 
 *  Arguments
 *    addr {IN}:	address.
 *
 *  returns
 *    address to the module.
 * 
 */

struct module * lkst_find_module_by_address(unsigned long addr)
{
	return (lkst_find_module_by_address_inline(addr));
}
EXPORT_SYMBOL(lkst_find_module_by_address);

#endif /* CONFIG_MODULES */

/*
 * Function lkst_maxvm_setup
 * This function gets size of lkst available memory area from kernel parameter.
 */

static int __init lkst_maxvm_setup(char *line)
{
	int size;
	if (get_option(&line, &size))
		lkst.available_memory_size = (size_t) size;
	return 1;
}
 
__setup("lkst_maxvm=", lkst_maxvm_setup);


/*
 * Function lkst_dummy_vmalloc
 * This function allocates dummy memory area to gets page directory entry
 * for LKST using memory area.
 * Ths area size is set by kernel parameter `lkst_maxvm=`. If omitted,
 * the size is set to vmalloc area size. But if 1/8 of physical memory size
 * is less than vmalloc area size, 1/8 of physical memory is allocated.
 * The allocation limit address is stored as lkst.limit_addr.
 * Note. The vmalloc function allocates required size + a page size.
 */

__inline__ void lkst_dummy_vmalloc(void)
{
	char *buf;
	unsigned long size;
	int i;
	size = lkst.available_memory_size;

	if (!size) {
		if ( (num_physpages >> 3) >
		     LKST_AVAILABLE_MEMORY_SIZE / PAGE_SIZE ) {
			size = LKST_AVAILABLE_MEMORY_SIZE;
		} else
			size = PAGE_SIZE * (num_physpages >> 3);
	}
	buf = vmalloc(size);

	if (buf==NULL) {
		printk(KERN_ERR "LKST warning: Cannot get available memory area(size=%ld) for LKST! Please restart kernel, or cannot allocate memory for LKST.\n", size);
		return;
	} 

	/* get pages dirty... */
	for ( i = 0; i < size ; i+= PAGE_SIZE) {
	  buf[i] = i+1;
	}
	vfree(buf);
	lkst.available_memory_size = size;
	(char *) lkst.limit_addr = (char *) buf + size;
	printk(KERN_INFO "LKST information: LKST Avaliable memory area is up to an address 0x%p, size is %ld bytes\n", lkst.limit_addr, size);
}

void lkst_owner_wakeup(int cpu)
{
	if(lkst.owner_ready[cpu]) {
		lkst.owner_ready[cpu]=0;
		wake_up_interruptible(&lkst.buffer_wq[cpu]);
	}
}
EXPORT_SYMBOL(lkst_owner_wakeup);

struct lkst_hook_info lkst_hook_info_table[LKST_MASKSET_TABLE_LEN_MAX];
EXPORT_SYMBOL(lkst_hook_info_table);

/* Declaration of hook header */
#define LKST_ETYPE_DECLARE_HOOKHEADER
#include <linux/lkst_events.h>
#undef  LKST_ETYPE_DECLARE_HOOKHEADER

void lkst_trc_hook_init(void)
{
	int ret, retall = 0;
	int i;

	/* Initialize lkst_hook_info_table */
	for(i=0; i<LKST_MASKSET_TABLE_LEN_MAX; i++){
		lkst_hook_info_table[i].hook_id = NULL;
		lkst_hook_info_table[i].phookrec = NULL;
	}

/* Initialization of hook header */
#define LKST_ETYPE_INIT_HOOK
#include <linux/lkst_events.h>
#undef  LKST_ETYPE_INIT_HOOK

	if (retall) {
		printk(KERN_ERR "LKST error(hook_init): hook_init failed: %u\n", ret);
	}
}

void lkst_trc_hook_cleanup(void)
{
	int ret, retall = 0;

/* Cleanup of hook */
#define LKST_ETYPE_TERMINATE_HOOK
#include <linux/lkst_events.h>
#undef  LKST_ETYPE_TERMINATE_HOOK

	if (retall) {
		printk(KERN_ERR "LKST error(hook_cleanup): hook_cleanup failed: %u\n", ret);
	}
}
EXPORT_SYMBOL(lkst_trc_hook_init);
EXPORT_SYMBOL(lkst_trc_hook_cleanup);
